import string

import numpy as np
import pandas as pd
import pytest

from plotnine import (
    aes,
    after_stat,
    geom_bar,
    geom_boxplot,
    geom_col,
    geom_jitter,
    geom_point,
    geom_rect,
    geom_text,
    ggplot,
    position_dodge,
    position_dodge2,
    position_jitter,
    position_jitterdodge,
    position_nudge,
    position_stack,
    scale_y_log10,
)
from plotnine.exceptions import PlotnineError
from plotnine.positions.position import position

n = 6
m = 10
random_state = np.random.RandomState(1234567890)
df1 = pd.DataFrame({"x": [1, 2, 1, 2], "y": [1, 1, 2, 2]})
df2 = pd.DataFrame(
    {
        "x": np.repeat(range(n + 1), range(n + 1)),
        "z": np.repeat(range(n // 2), range(3, n * 2, 4)),
    }
)
df3 = pd.DataFrame(
    {
        "x": random_state.choice(["A", "B"], n * m),
        "y": random_state.randint(0, 20, n * m),
        "c": random_state.choice([False, False, True, False], n * m),
    }
)
random_state.seed(1234567890)


def test_jitter():
    df1 = pd.DataFrame({"x": [1, 2, 1, 2], "y": [1, 1, 2, 2]})
    p = (
        ggplot(df1, aes("x", "y"))
        + geom_point(size=10)
        + geom_jitter(size=10, color="red", random_state=random_state)
        + geom_jitter(
            size=10,
            color="blue",
            width=0.1,
            height=0.1,
            random_state=random_state,
        )
    )
    assert p == "jitter"

    with pytest.raises(PlotnineError):
        geom_jitter(position=position_jitter(), width=0.1)


def test_nudge():
    p = (
        ggplot(df1, aes("x", "y"))
        + geom_point(size=10)
        + geom_point(size=10, color="red", position=position_nudge(0.25, 0.25))
    )
    assert p == "nudge"


def test_stack():
    p = ggplot(df2, aes("factor(z)")) + geom_bar(
        aes(fill="factor(x)"), position="stack"
    )
    assert p == "stack"


def test_stack_negative():
    df = df1.copy()
    _loc = df.columns.get_loc
    df.iloc[0, _loc("y")] *= -1
    df.iloc[len(df) - 1, _loc("y")] *= -1
    p = (
        ggplot(df)
        + geom_col(aes("factor(x)", "y", fill="factor(y)"), position="stack")
        + geom_text(
            aes("factor(x)", "y", label="y"),
            position=position_stack(vjust=0.5),
        )
    )

    assert p == "stack-negative"


def test_stack_non_linear_scale():
    df = pd.DataFrame(
        {
            "x": "x",
            "value": [0.1, 10, 100, 1000],
            "cat": ["small", "small", "big", "big"],
        }
    )

    p = (
        ggplot(df, aes("x", "value", fill="cat"))
        + geom_col()
        + scale_y_log10()
    )
    assert p == "stack-non-linear-scale"


def test_fill():
    p = ggplot(df2, aes("factor(z)")) + geom_bar(
        aes(fill="factor(x)"), position="fill"
    )
    assert p == "fill"


def test_dodge():
    p = ggplot(df2, aes("factor(z)")) + geom_bar(
        aes(fill="factor(x)"), position="dodge"
    )
    assert p == "dodge"


def test_dodge_preserve_single():
    df1 = pd.DataFrame({"x": ["a", "b", "b"], "y": ["a", "a", "b"]})
    p = ggplot(df1, aes("x", fill="y")) + geom_bar(
        position=position_dodge(preserve="single")
    )
    assert p == "dodge_preserve_single"


def test_dodge_preserve_single_text():
    df1 = pd.DataFrame({"x": ["a", "b", "b", "b"], "y": ["a", "a", "b", "b"]})

    d = position_dodge(preserve="single", width=0.9)
    p = (
        ggplot(df1, aes("x", fill="y"))
        + geom_bar(position=d)
        + geom_text(
            aes(y=after_stat("count"), label=after_stat("count")),
            stat="count",
            position=d,
            va="bottom",
        )
    )
    assert p == "dodge_preserve_single_text"


def test_dodge2():
    p = ggplot(df3, aes("x", "y", color="c")) + geom_boxplot(
        position="dodge2", size=2
    )
    assert p == "dodge2"


def test_dodge2_varwidth():
    p = ggplot(df3, aes("x", "y", color="c")) + geom_boxplot(
        position=position_dodge2(preserve="single"), varwidth=True, size=2
    )
    assert p == "dodge2_varwidth"


def test_dodge2_preserve_single_interval():
    n = 3
    df = pd.DataFrame({"x": range(1, n + 1), "y": range(1, n + 1)})

    p = ggplot(
        df, aes(xmin="x-.45", xmax="x+.45", ymin=0, ymax="y")
    ) + geom_rect(position=position_dodge2(preserve="single"))
    assert p == "dodge2_preserve_single_interval"


def test_jitterdodge():
    df = pd.DataFrame(
        {
            "x": np.ones(n * 2),
            "y": np.repeat(np.arange(n), 2),
            "letters": np.repeat(list(string.ascii_lowercase[:n]), 2),
        }
    )
    position = position_jitterdodge(random_state=random_state)

    p = (
        ggplot(df, aes("x", "y", fill="letters"))
        + geom_point(size=10, fill="black")
        + geom_point(size=10, position=position)
    )
    assert p == "jitterdodge"


def test_position_from_geom():
    geom = geom_point(position="jitter")
    assert isinstance(position.from_geom(geom), position_jitter)

    geom = geom_point(position="position_jitter")
    assert isinstance(position.from_geom(geom), position_jitter)

    geom = geom_point(position=position_jitter())
    assert isinstance(position.from_geom(geom), position_jitter)

    geom = geom_point(position=position_jitter)
    assert isinstance(position.from_geom(geom), position_jitter)


def test_dodge_empty_data():
    empty_df = pd.DataFrame({"x": [], "y": []})
    p = (
        ggplot(df1, aes("x", "y"))
        + geom_point()
        + geom_rect(
            empty_df,
            aes(xmin="x", xmax="x+1", ymin="y", ymax="y+1"),
            position="dodge",
        )
    )
    p.draw_test()



"""
Routines to adjust subplot params so that subplots are
nicely fit in the figure. In doing so, only axis labels, tick labels, axes
titles and offsetboxes that are anchored to axes are currently considered.

Internally, this module assumes that the margins (left margin, etc.) which are
differences between ``Axes.get_tightbbox`` and ``Axes.bbox`` are independent of
Axes position. This may fail if ``Axes.adjustable`` is ``datalim`` as well as
such cases as when left or right margin are affected by xlabel.
"""
from __future__ import annotations

import typing
from copy import deepcopy
from dataclasses import dataclass, fields
from itertools import chain

from matplotlib._tight_layout import get_subplotspec_list
from matplotlib.offsetbox import AnchoredOffsetbox

from ..facets import facet_grid, facet_null, facet_wrap
from .patches import SFancyBboxPatch
from .utils import bbox_in_figure_space, tight_bbox_in_figure_space

if typing.TYPE_CHECKING:
    from dataclasses import Field
    from typing import Generator, Iterator, Literal, TypeAlias

    from plotnine.typing import (
        Artist,
        Axes,
        Facet,
        Figure,
        LegendPosition,
        Text,
        XTick,
        YTick,
    )

    from .layout_engine import LayoutPack

    AxesLocation: TypeAlias = Literal[
        "all", "first_row", "last_row", "first_col", "last_col"
    ]

# Note
# Margin around the plot are specfied in figure coordinates
# We interprete that value to be a fraction of the width. So along
# the vertical direction we multiply by W/H to get equal space
# in both directions


@dataclass
class GridSpecParams:
    """
    Gridspec Parameters
    """

    left: float
    right: float
    top: float
    bottom: float
    wspace: float
    hspace: float


@dataclass
class _side_spaces:
    """
    Base class to for spaces

    A *_space class should track the size taken up by all the objects that
    may fall on that side of the panel. The same name may appear in multiple
    side classes (e.g. legend), but atmost only one of those parameters will
    have a non-zero value.
    """

    @property
    def total(self) -> float:
        """
        Total space
        """
        return sum(getattr(self, f.name) for f in fields(self))

    def sum_upto(self, item: str) -> float:
        """
        Sum of space upto but not including item

        Sums starting at the edge of the figure i.e. the "plot_margin".
        """

        def _fields_upto(x: _side_spaces) -> Generator[Field, None, None]:
            for f in fields(x):
                if f.name == item:
                    break
                yield f

        return sum(getattr(self, f.name) for f in _fields_upto(self))


@dataclass
class left_spaces(_side_spaces):
    """
    Space in the figure for artists on the left of the panels

    Ordered from the edge of the figure and going inwards
    """

    plot_margin: float = 0
    legend: float = 0
    legend_box_spacing: float = 0
    axis_title_y: float = 0
    axis_title_y_margin_right: float = 0
    axis_y: float = 0

    def edge(self, item: str) -> float:
        """
        Distance w.r.t figure width from the left edge of the figure
        """
        return self.sum_upto(item)


@dataclass
class right_spaces(_side_spaces):
    """
    Space in the figure for artists on the right of the panels

    Ordered from the edge of the figure and going inwards
    """

    plot_margin: float = 0
    legend: float = 0
    legend_box_spacing: float = 0
    right_strip_width: float = 0

    def edge(self, item: str) -> float:
        """
        Distance w.r.t figure width from the right edge of the figure
        """
        return 1 - self.sum_upto(item)


@dataclass
class top_spaces(_side_spaces):
    """
    Space in the figure for artists above the panels

    Ordered from the edge of the figure and going inwards
    """

    plot_margin: float = 0
    plot_title: float = 0
    plot_title_margin_bottom: float = 0
    plot_subtitle: float = 0
    plot_subtitle_margin_bottom: float = 0
    legend: float = 0
    legend_box_spacing: float = 0
    top_strip_height: float = 0

    def edge(self, item: str) -> float:
        """
        Distance w.r.t figure height from the top edge of the figure
        """
        return 1 - self.sum_upto(item)


@dataclass
class bottom_spaces(_side_spaces):
    """
    Space in the figure for artists below the panels

    Ordered from the edge of the figure and going inwards
    """

    plot_margin: float = 0
    plot_caption: float = 0
    plot_caption_margin_top: float = 0
    legend: float = 0
    legend_box_spacing: float = 0
    axis_title_x: float = 0
    axis_title_x_margin_top: float = 0
    axis_x: float = 0

    def edge(self, item: str) -> float:
        """
        Distance w.r.t figure height from the bottom edge of the figure
        """
        return self.sum_upto(item)


@dataclass
class LRTBSpaces:
    """
    Space for components in all directions around the panels
    """

    l: left_spaces
    r: right_spaces
    t: top_spaces
    b: bottom_spaces

    @property
    def left(self):
        """
        Left of the panels in figure space
        """
        return self.l.total

    @property
    def right(self):
        """
        Right of the panels in figure space
        """
        return 1 - self.r.total

    @property
    def top(self):
        """
        Top of the panels in figure space
        """
        return 1 - self.t.total

    @property
    def bottom(self):
        """
        Bottom of the panels in figure space
        """
        return self.b.total


@dataclass
class WHSpaceParts:
    """
    Width-Height Spaces

    We need these in one places for easy access
    """

    W: float  # Figure width
    H: float  # Figure height
    w: float  # Axes width w.r.t figure in [0, 1]
    h: float  # Axes height w.r.t figure in [0, 1]
    sw: float  # horizontal spacing btn panels w.r.t figure
    sh: float  # vertical spacing btn panels w.r.t figure
    wspace: float  # mpl.subplotpars.wspace
    hspace: float  # mpl.subplotpars.hspace


@dataclass
class TightParams:
    """
    All parameters computed for the plotnine tight layout engine
    """

    spaces: LRTBSpaces
    grid: GridSpecParams

    def __init__(self, spaces: LRTBSpaces, wspace: float, hspace: float):
        self.spaces = spaces
        self.grid = GridSpecParams(
            left=spaces.left,
            right=spaces.right,
            top=spaces.top,
            bottom=spaces.bottom,
            wspace=wspace,
            hspace=hspace,
        )

    def to_aspect_ratio(
        self, facet: Facet, ratio: float, parts: WHSpaceParts
    ) -> TightParams:
        """
        Modify TightParams to get a given aspect ratio
        """
        current_ratio = (parts.h * parts.H) / (parts.w * parts.W)
        increase_aspect_ratio = ratio > current_ratio
        if increase_aspect_ratio:  # Taller panel
            return self._reduce_width(facet, ratio, parts)
        else:
            return self._reduce_height(facet, ratio, parts)

    def _reduce_height(
        self, facet: Facet, ratio: float, parts: WHSpaceParts
    ) -> TightParams:
        """
        Reduce the height of axes to get the aspect ratio
        """
        params = deepcopy(self)
        spaces = params.spaces
        grid = params.grid

        # New height w.r.t figure height
        h1 = ratio * parts.w * (parts.W / parts.H)

        # Half of the total vertical reduction w.r.t figure height
        dh = (parts.h - h1) * facet.nrow / 2

        # Reduce plot area height
        grid.top -= dh
        grid.bottom += dh
        grid.hspace = parts.sh / h1

        # Add more vertical plot margin
        spaces.t.plot_margin += dh
        spaces.b.plot_margin += dh
        return params

    def _reduce_width(
        self, facet: Facet, ratio: float, parts: WHSpaceParts
    ) -> TightParams:
        """
        Reduce the width of axes to get the aspect ratio
        """
        params = deepcopy(self)
        spaces = params.spaces
        grid = params.grid

        # New width w.r.t figure width
        w1 = (parts.h * parts.H) / (ratio * parts.W)

        # Half of the total horizontal reduction w.r.t figure width
        dw = (parts.w - w1) * facet.ncol / 2

        # Reduce width
        grid.left += dw
        grid.right -= dw
        grid.wspace = parts.sw / w1

        # Add more horizontal margin
        spaces.l.plot_margin += dw
        spaces.r.plot_margin += dw
        return params


def get_plotnine_tight_layout(pack: LayoutPack) -> TightParams:
    """
    Compute tight layout parameters
    """
    spaces = LRTBSpaces(
        l=calculate_left_spaces(pack),
        r=calculate_right_spaces(pack),
        t=calculate_top_spaces(pack),
        b=calculate_bottom_spaces(pack),
    )
    parts = calculate_panel_spacing(pack, spaces)
    params = TightParams(spaces, parts.wspace, parts.hspace)
    ratio = pack.facet._aspect_ratio()
    if ratio is not None:
        params = params.to_aspect_ratio(pack.facet, ratio, parts)
    return params


def calculate_panel_spacing(
    pack: LayoutPack, spaces: LRTBSpaces
) -> WHSpaceParts:
    """
    Spacing between the panels (wspace & hspace)

    Both spaces are calculated from a fraction of the width.
    This ensures that the same fraction gives equals space
    in both directions.
    """
    if isinstance(pack.facet, facet_wrap):
        return _calculate_panel_spacing_facet_wrap(pack, spaces)
    elif isinstance(pack.facet, facet_grid):
        return _calculate_panel_spacing_facet_grid(pack, spaces)
    elif isinstance(pack.facet, facet_null):
        return _calculate_panel_spacing_facet_null(pack, spaces)
    return WHSpaceParts(0, 0, 0, 0, 0, 0, 0, 0)


def _calculate_panel_spacing_facet_grid(
    pack: LayoutPack, spaces: LRTBSpaces
) -> WHSpaceParts:
    """
    Calculate spacing parts for facet_grid
    """
    _property = pack.theme.themeables.property

    ncol = pack.facet.ncol
    nrow = pack.facet.nrow

    W, H = _property("figure_size")

    # Both spacings are specified as fractions of the figure width
    sw = _property("panel_spacing_x")
    sh = _property("panel_spacing_y") * W / H

    # width and height of axes as fraction of figure width & heigt
    w = ((spaces.right - spaces.left) - sw * (ncol - 1)) / ncol
    h = ((spaces.top - spaces.bottom) - sh * (nrow - 1)) / nrow

    # Spacing as fraction of axes width & height
    wspace = sw / w
    hspace = sh / h

    return WHSpaceParts(W, H, w, h, sw, sh, wspace, hspace)


def _calculate_panel_spacing_facet_wrap(
    pack: LayoutPack, spaces: LRTBSpaces
) -> WHSpaceParts:
    """
    Calculate spacing parts for facet_wrap
    """
    _property = pack.theme.themeables.property

    ncol = pack.facet.ncol
    nrow = pack.facet.nrow

    W, H = _property("figure_size")

    # Both spacings are specified as fractions of the figure width
    sw = _property("panel_spacing_x")
    sh = _property("panel_spacing_y") * W / H

    # A fraction of the strip height
    # Effectively slides the strip
    #   +ve: Away from the panel
    #    0:  Top of the panel
    #   -ve: Into the panel
    # Where values <= -1, put the strip completly into
    # the panel. We do not worry about larger -ves.
    strip_align_x = _property("strip_align_x")

    # Only interested in the proportion of the strip that
    # does not overlap with the panel
    if strip_align_x > -1:
        sh += spaces.t.top_strip_height * (1 + strip_align_x)

    if isinstance(pack.facet, facet_wrap):
        if pack.facet.free["x"]:
            sh += get_max_xaxis_height(pack)
        if pack.facet.free["y"]:
            sw += get_max_yaxis_width(pack)

    # width and height of axes as fraction of figure width & heigt
    w = ((spaces.right - spaces.left) - sw * (ncol - 1)) / ncol
    h = ((spaces.top - spaces.bottom) - sh * (nrow - 1)) / nrow

    # Spacing as fraction of axes width & height
    wspace = sw / w
    hspace = sh / h

    return WHSpaceParts(W, H, w, h, sw, sh, wspace, hspace)


def _calculate_panel_spacing_facet_null(
    pack: LayoutPack, spaces: LRTBSpaces
) -> WHSpaceParts:
    """
    Calculate spacing parts for facet_null
    """
    _property = pack.theme.themeables.property
    W, H = pack.theme.themeables.property("figure_size")
    w = spaces.right - spaces.left
    h = spaces.top - spaces.bottom
    return WHSpaceParts(W, H, w, h, 0, 0, 0, 0)


def calculate_left_spaces(pack: LayoutPack) -> left_spaces:
    """
    Left margin
    """
    _property = pack.theme.themeables.property
    space = left_spaces(plot_margin=_property("plot_margin"))

    if pack.legend and pack.legend_position == "left":
        space.legend += bbox_in_figure_space(
            pack.legend, pack.figure, pack.renderer
        ).width
        space.legend_box_spacing = _property("legend_box_spacing")

    if pack.axis_title_y:
        space.axis_title_y_margin_right = _property(
            "axis_title_y", "margin"
        ).get_as("r", "fig")
        space.axis_title_y = bbox_in_figure_space(
            pack.axis_title_y, pack.figure, pack.renderer
        ).width

    # Account for the space consumed by the axis even if it comes
    # after the axis text
    space.axis_y = get_max_yaxis_width(pack, "first_col")
    return space


def calculate_right_spaces(pack: LayoutPack) -> right_spaces:
    """
    Calculate Space for artists on the right of the panel
    """
    _property = pack.theme.themeables.property
    space = right_spaces(plot_margin=_property("plot_margin"))

    if pack.legend and pack.legend_position == "right":
        space.legend = bbox_in_figure_space(
            pack.legend, pack.figure, pack.renderer
        ).width
        space.legend_box_spacing = _property("legend_box_spacing")

    space.right_strip_width = get_right_strip_width_in_last_col(pack)
    return space


def calculate_top_spaces(pack: LayoutPack) -> top_spaces:
    """
    Calculate space for artists above the panels
    """
    _property = pack.theme.themeables.property

    W, H = _property("figure_size")
    F = W / H
    space = top_spaces(plot_margin=_property("plot_margin") * F)

    if pack.plot_title:
        space.plot_title = bbox_in_figure_space(
            pack.plot_title, pack.figure, pack.renderer
        ).height
        space.plot_title_margin_bottom = (
            _property("plot_title", "margin").get_as("b", "fig") * F
        )

    if pack.plot_subtitle:
        space.plot_subtitle = bbox_in_figure_space(
            pack.plot_subtitle, pack.figure, pack.renderer
        ).height
        space.plot_subtitle_margin_bottom = (
            _property("plot_subtitle", "margin").get_as("b", "fig") * F
        )

    if pack.legend and pack.legend_position == "top":
        space.legend = bbox_in_figure_space(
            pack.legend, pack.figure, pack.renderer
        ).height
        space.legend_box_spacing = _property("legend_box_spacing") * F

    space.top_strip_height = get_top_strip_height_in_first_row(pack)
    return space


def calculate_bottom_spaces(pack: LayoutPack) -> bottom_spaces:
    """
    Calculate space for the artists below the panels
    """
    _property = pack.theme.themeables.property

    W, H = pack.theme.themeables.property("figure_size")
    F = W / H
    space = bottom_spaces(plot_margin=_property("plot_margin") * F)

    if pack.plot_caption:
        space.plot_caption = bbox_in_figure_space(
            pack.plot_caption, pack.figure, pack.renderer
        ).height
        space.plot_caption_margin_top = (
            _property("plot_caption", "margin").get_as("t", "fig") * F
        )

    if pack.legend and pack.legend_position == "bottom":
        space.legend = bbox_in_figure_space(
            pack.legend, pack.figure, pack.renderer
        ).height
        space.legend_box_spacing = _property("legend_box_spacing") * F

    if pack.axis_title_x:
        space.axis_title_x = bbox_in_figure_space(
            pack.axis_title_x, pack.figure, pack.renderer
        ).height
        space.axis_title_x_margin_top = (
            _property("axis_title_x", "margin").get_as("t", "fig") * F
        )

    # Account for the space consumed by the axis
    space.axis_x = get_max_xaxis_height(pack, "last_row")
    return space


def filter_axes(axs, get: AxesLocation = "all"):
    """
    Return subset of axes
    """
    if get == "all":
        return axs

    # e.g. is_first_row, is_last_row, ..
    pred_method = f"is_{get}"
    return [
        ax
        for spec, ax in zip(get_subplotspec_list(axs), axs)
        if getattr(spec, pred_method)()
    ]


def is_top_strip_boxpatch(artist: Artist) -> bool:
    """
    Return True if artist is a patch/background of the top strip of a facet
    """
    if isinstance(artist, SFancyBboxPatch):
        return artist.position == "top"
    return False


def is_right_strip_boxpatch(artist: Artist) -> bool:
    """
    Return True if artist is a patch/background of the right strip of a facet
    """
    if isinstance(artist, SFancyBboxPatch):
        return artist.position == "right"
    return False


def get_top_strip_boxpatches_in_first_row(
    axs: list[Axes],
) -> list[SFancyBboxPatch]:
    """
    Return all box patches on the top of the first row
    """
    return [
        child
        for ax in filter_axes(axs, "first_row")
        for child in ax.get_children()
        if is_top_strip_boxpatch(child)
    ]


def get_right_strip_boxpatches_in_last_col(
    axs: list[Axes],
) -> list[SFancyBboxPatch]:
    """
    Return all box patches on the right of the last column
    """
    return [
        child
        for ax in filter_axes(axs, "last_col")
        for child in ax.get_children()
        if is_right_strip_boxpatch(child)
    ]


def get_top_strip_height_in_first_row(pack: LayoutPack) -> float:
    """
    Return the height of the top strip
    """
    patches = get_top_strip_boxpatches_in_first_row(pack.axs)
    heights = [
        bbox_in_figure_space(patch, pack.figure, pack.renderer).height
        for patch in patches
    ]
    return max(heights) if len(heights) else 0


def get_right_strip_width_in_last_col(pack: LayoutPack) -> float:
    """
    Return the width of the right strip
    """
    patches = get_right_strip_boxpatches_in_last_col(pack.axs)
    widths = [
        bbox_in_figure_space(patch, pack.figure, pack.renderer).width
        for patch in patches
    ]
    return max(widths) if len(widths) else 0


def get_xticks(ax: Axes) -> Iterator[XTick]:
    """
    Return all XTicks
    """
    return chain(ax.xaxis.get_major_ticks(), ax.xaxis.get_minor_ticks())


def get_yticks(ax: Axes) -> Iterator[YTick]:
    """
    Return all YTicks
    """
    return chain(ax.yaxis.get_major_ticks(), ax.yaxis.get_minor_ticks())


def get_max_xaxis_height(
    pack: LayoutPack,
    axes_loc: AxesLocation = "all",
) -> float:
    """
    Return maximum height[inches] of an xaxis
    """
    H = pack.figure.get_figheight()
    # Note: Using the bbox of for axis ignores the ticklines
    heights = [
        tight_bbox_in_figure_space(
            tick.tick1line, pack.figure, pack.renderer
        ).height
        + tick.get_pad() / (72 * H)
        + tight_bbox_in_figure_space(
            tick.label1, pack.figure, pack.renderer
        ).height
        for ax in filter_axes(pack.axs, axes_loc)
        for tick in get_xticks(ax)
    ]
    return max(heights) if len(heights) else 0


def get_max_yaxis_width(
    pack: LayoutPack,
    axes_loc: AxesLocation = "all",
) -> float:
    """
    Return maximum width[inches] of an xaxis
    """
    W = pack.figure.get_figwidth()
    # Note: Using the bbox of for axis ignores the ticklines
    widths = [
        tight_bbox_in_figure_space(
            tick.tick1line, pack.figure, pack.renderer
        ).width
        + tick.get_pad() / (72 * W)
        + tight_bbox_in_figure_space(
            tick.label1, pack.figure, pack.renderer
        ).width
        for ax in filter_axes(pack.axs, axes_loc)
        for tick in get_yticks(ax)
    ]
    return max(widths) if len(widths) else 0


def set_figure_artist_positions(
    pack: LayoutPack,
    tparams: TightParams,
):
    """
    Set the x,y position of the artists around the panels
    """
    _property = pack.theme.themeables.property
    spaces = tparams.spaces
    grid = tparams.grid

    if pack.plot_title:
        try:
            ha = _property("plot_title", "ha")
        except KeyError:
            ha = "center"
        pack.plot_title.set_y(spaces.t.edge("plot_title"))
        horizonally_align_text_with_panels(pack.plot_title, grid, ha)

    if pack.plot_subtitle:
        try:
            ha = _property("plot_subtitle", "ha")
        except KeyError:
            ha = "center"
        pack.plot_subtitle.set_y(spaces.t.edge("plot_subtitle"))
        horizonally_align_text_with_panels(pack.plot_subtitle, grid, ha)

    if pack.plot_caption:
        try:
            ha = _property("plot_caption", "ha")
        except KeyError:
            ha = "right"
        pack.plot_caption.set_y(spaces.b.edge("plot_caption"))
        horizonally_align_text_with_panels(pack.plot_caption, grid, ha)

    if pack.axis_title_x:
        try:
            ha = _property("axis_title_x", "ha")
        except KeyError:
            ha = "center"
        pack.axis_title_x.set_y(spaces.b.edge("axis_title_x"))
        horizonally_align_text_with_panels(pack.axis_title_x, grid, ha)

    if pack.axis_title_y:
        try:
            va = _property("axis_title_y", "va")
        except KeyError:
            va = "center"
        pack.axis_title_y.set_x(spaces.l.edge("axis_title_y"))
        vertically_align_text_with_panels(pack.axis_title_y, grid, va)

    if pack.legend and pack.legend_position:
        set_legend_position(
            pack.legend, pack.legend_position, tparams, pack.figure
        )


def horizonally_align_text_with_panels(
    text: Text, grid: GridSpecParams, ha: str
):
    """
    Horizontal justification

    Reinterpret horizontal alignment to be justification about the panels.
    """
    if ha == "center":
        text.set_x((grid.left + grid.right) / 2)
    elif ha == "left":
        text.set_x(grid.left)
    elif ha == "right":
        text.set_x(grid.right)


def vertically_align_text_with_panels(
    text: Text, grid: GridSpecParams, va: str
):
    """
    Vertical justification

    Reinterpret vertical alignment to be justification about the panels.
    """
    if va == "center":
        text.set_y((grid.top + grid.bottom) / 2)
    elif va == "top":
        text.set_y(grid.top)
    elif va == "bottom":
        text.set_y(grid.bottom)


def set_legend_position(
    legend: AnchoredOffsetbox,
    position: LegendPosition,
    tparams: TightParams,
    fig: Figure,
):
    """
    Place legend and align it centerally with respect to the panels
    """
    grid = tparams.grid
    spaces = tparams.spaces

    if position in ("right", "left"):
        y = (grid.top + grid.bottom) / 2
        if position == "left":
            x = spaces.l.edge("legend")
            loc = "center left"
        else:
            x = spaces.r.edge("legend")
            loc = "center right"
    elif position in ("top", "bottom"):
        x = (grid.right + grid.left) / 2
        if position == "top":
            y = spaces.t.edge("legend")
            loc = "upper center"
        else:
            y = spaces.b.edge("legend")
            loc = "lower center"
    else:
        x, y = position
        loc = "center"

    anchor_point = (x, y)
    legend.loc = AnchoredOffsetbox.codes[loc]
    legend.set_bbox_to_anchor(anchor_point, fig.transFigure)



from __future__ import annotations

import re
import typing
from functools import lru_cache
from textwrap import dedent, indent, wrap

import numpy as np

if typing.TYPE_CHECKING:
    from typing import Any, Type, TypeVar

    from plotnine.typing import Geom, Scale, Stat

    T = TypeVar("T")


# Parameter arguments that are listed first in the geom and
# stat class signatures

common_geom_params = [
    "mapping",
    "data",
    "stat",
    "position",
    "na_rm",
    "inherit_aes",
    "show_legend",
    "raster",
]
common_geom_param_values = {
    "mapping": None,
    "data": None,
    "inherit_aes": True,
    "show_legend": None,
    "raster": False,
}

common_stat_params = ["mapping", "data", "geom", "position", "na_rm"]
common_stat_param_values = common_geom_param_values

# Templates for docstrings

GEOM_SIGNATURE_TPL = """
.. rubric:: Usage

::

    {signature}

Only the ``data`` and ``mapping`` can be positional, the rest must
be keyword arguments. ``**kwargs`` can be aesthetics (or parameters)
used by the ``stat``.
"""

AESTHETICS_TABLE_TPL = """
{table}

The **bold** aesthetics are required.
"""

STAT_SIGNATURE_TPL = """
.. rubric:: Usage

::

    {signature}

Only the ``mapping`` and ``data`` can be positional, the rest must
be keyword arguments. ``**kwargs`` can be aesthetics (or parameters)
used by the ``geom``.
"""


common_params_doc = {
    "mapping": """\
Aesthetic mappings created with :meth:`~plotnine.aes`. If specified and \
:py:`inherit.aes=True`, it is combined with the default mapping for the plot. \
You must supply mapping if there is no plot mapping.""",
    "data": """\
The data to be displayed in this layer. If :py:`None`, the data from \
from the :py:`ggplot()` call is used. If specified, it overrides the \
data from the :py:`ggplot()` call.""",
    "stat": """\
The statistical transformation to use on the data for this layer. \
If it is a string, it must be the registered and known to Plotnine.""",
    "position": """\
Position adjustment. If it is a string, it must be registered and \
known to Plotnine.""",
    "na_rm": """\
If :py:`False`, removes missing values with a warning. If :py:`True` \
silently removes missing values.""",
    "inherit_aes": """\
If :py:`False`, overrides the default aesthetics.""",
    "show_legend": """\
Whether this layer should be included in the legends. :py:`None` the \
default, includes any aesthetics that are mapped. If a :class:`bool`, \
:py:`False` never includes and :py:`True` always includes. A \
:class:`dict` can be used to *exclude* specific aesthetis of the layer \
from showing in the legend. e.g :py:`show_legend={'color': False}`, \
any other aesthetic are included by default.""",
    "raster": """\
If ``True``, draw onto this layer a raster (bitmap) object even if\
the final image is in vector format.""",
}


GEOM_PARAMS_TPL = """\
mapping : aes, optional
    {mapping}
    {_aesthetics_doc}
data : dataframe, optional
    {data}
stat : str or stat, optional (default: {default_stat})
    {stat}
position : str or position, optional (default: {default_position})
    {position}
na_rm : bool, optional (default: {default_na_rm})
    {na_rm}
inherit_aes : bool, optional (default: {default_inherit_aes})
    {inherit_aes}
show_legend : bool or dict, optional (default: None)
    {show_legend}
raster : bool, optional (default: {default_raster})
    {raster}
"""

STAT_PARAMS_TPL = """\
mapping : aes, optional
    {mapping}
    {_aesthetics_doc}
data : dataframe, optional
    {data}
geom : str or geom, optional (default: {default_geom})
    {stat}
position : str or position, optional (default: {default_position})
    {position}
na_rm : bool, optional (default: {default_na_rm})
    {na_rm}
"""

DOCSTRING_SECTIONS = {
    "parameters",
    "see also",
    "note",
    "notes",
    "example",
    "examples",
}

PARAM_PATTERN = re.compile(r"\s*" r"([_A-Za-z]\w*)" r"\s:\s")


def dict_to_table(header: tuple[str, str], contents: dict[str, str]) -> str:
    """
    Convert dict to an (n x 2) table

    Parameters
    ----------
    header : tuple
        Table header. Should have a length of 2.
    contents : dict
        The key becomes column 1 of table and the
        value becomes column 2 of table.

    Examples
    --------
    >>> d = {"alpha": 1, "color": "blue", "fill": None}
    >>> print(dict_to_table(("Aesthetic", "Default Value"), d))
    ========= =========
    Aesthetic Default Value
    ========= =========
    alpha     :py:`1`
    color     :py:`'blue'`
    fill      :py:`None`
    ========= =========
    """

    def to_text(row: tuple[str, str]) -> str:
        name, value = row
        m = max_col1_size + 1 - len(name)
        spacing = " " * m
        return "".join([name, spacing, value])

    def longest_value(row: tuple[str, str]) -> int:
        return max(len(value) for value in row)

    rows = []
    for name, value in contents.items():
        # code highlighting
        if value != "":
            if isinstance(value, str):
                value = f"'{value}'"
            value = f":py:`{value}`"
        rows.append((name, value))

    n = max(longest_value(row) for row in [header] + rows)
    hborder = "=" * n, "=" * n
    rows = [hborder, header, hborder] + rows + [hborder]
    max_col1_size = np.max([len(col1) for col1, _ in rows])
    table = "\n".join([to_text(row) for row in rows])
    return table


def make_signature(
    name: str,
    params: dict[str, Any],
    common_params: list[str],
    common_param_values: dict[str, Any],
) -> str:
    """
    Create a signature for a geom or stat

    Gets the DEFAULT_PARAMS (params) and creates are comma
    separated list of the `name=value` pairs. The common_params
    come first in the list, and they get take their values from
    either the params-dict or the common_geom_param_values-dict.
    """
    tokens = []
    seen = set()

    def tokens_append(key: str, value: Any):
        if isinstance(value, str):
            value = f"'{value}'"
        tokens.append(f"{key}={value}")

    # preferred params come first
    for key in common_params:
        seen.add(key)
        try:
            value = params[key]
        except KeyError:
            value = common_param_values[key]
        tokens_append(key, value)

    # other params (these are the geom/stat specific parameters
    for key in set(params) - seen:
        tokens_append(key, params[key])

    # name, 1 opening bracket, 4 spaces in SIGNATURE_TPL
    s_params = ", ".join(tokens)
    s1 = f"{name}("
    s2 = f"{s_params}, **kwargs)"
    line_width = 78 - len(s1)
    indent_spaces = " " * (len(s1) + 4)
    s2_lines = wrap(s2, width=line_width)
    s2_indented = f"\n{indent_spaces}".join(s2_lines)
    return f"{s1}{s2_indented}"


@lru_cache(maxsize=256)
def docstring_section_lines(docstring: str, section_name: str) -> str:
    """
    Return a section of a numpydoc string

    Paramters
    ---------
    docstring : str
        Docstring
    section_name : str
        Name of section to return

    Returns
    -------
    section : str
        Section minus the header
    """
    lines = []
    inside_section = False
    underline = "-" * len(section_name)
    expect_underline = False
    for line in docstring.splitlines():
        _line = line.strip().lower()

        if expect_underline:
            expect_underline = False
            if _line == underline:
                inside_section = True
                continue

        if _line == section_name:
            expect_underline = True
        elif _line in DOCSTRING_SECTIONS:
            # next section
            break
        elif inside_section:
            lines.append(line)
    return "\n".join(lines)


def docstring_parameters_section(obj: Any) -> str:
    """
    Return the parameters section of a docstring
    """
    return docstring_section_lines(obj.__doc__, "parameters")


def param_spec(line: str) -> str | None:
    """
    Identify and return parameter

    Parameters
    ----------
    line : str
        A line in the parameter section.

    Returns
    -------
    name : str or None
        Name of the parameter if the line for the parameter
        type specification and None otherwise.

    Examples
    --------
    >>> param_spec('line : str')
    breaks
    >>> param_spec("    A line in the parameter section.")
    """
    m = PARAM_PATTERN.match(line)
    return m.group(1) if m else None


def parameters_str_to_dict(param_section: str) -> dict[str, str]:
    """
    Convert a param section to a dict

    Parameters
    ----------
    param_section : str
        Text in the parameter section

    Returns
    -------
    d : dict
        Dictionary of the parameters in the order that they
        are described in the parameters section. The dict
        is of the form ``{param: all_parameter_text}``.
        You can reconstruct the ``param_section`` from the
        keys of the dictionary.

    See Also
    --------
    :func:`parameters_dict_to_str`
    """
    d = {}
    previous_param = ""
    param_desc: list[str] = []
    for line in param_section.split("\n"):
        param = param_spec(line)
        if param:
            if previous_param:
                d[previous_param] = "\n".join(param_desc)
            param_desc = [line]
            previous_param = param
        elif param_desc:
            param_desc.append(line)

    if previous_param:
        d[previous_param] = "\n".join(param_desc)

    return d


def parameters_dict_to_str(d: dict[str, str]) -> str:
    """
    Convert a dict of param section to a string

    Parameters
    ----------
    d : dict
        Parameters and their descriptions in a docstring

    Returns
    -------
    param_section : str
        Text in the parameter section

    See Also
    --------
    :func:`parameters_str_to_dict`
    """
    return "\n".join(d.values())


def qualified_name(s: str | type | object, prefix: str) -> str:
    """
    Return the qualified name of s

    Only if s does not start with the prefix

    Examples
    --------
    >>> qualified_name('bin', 'stat_')
    '~plotnine.stats.stat_bin'
    >>> qualified_name('point', 'geom_')
    '~plotnine.geoms.geom_point'
    >>> qualified_name('stack', 'position_')
    '~plotnine.positions.position_'
    """
    lookup = {
        "stat_": "~plotnine.stats.stat_",
        "geom_": "~plotnine.geoms.geom_",
        "position_": "~plotnine.positions.position_",
    }
    if isinstance(s, str):
        if not s.startswith(prefix) and prefix in lookup:
            pre = lookup[prefix]
            s = f"{pre}{s}"
    elif isinstance(s, type):
        s = s.__name__
    else:
        s = s.__class__.__name__
    return s


def document_geom(geom: type[Geom]) -> type[Geom]:
    """
    Create a structured documentation for the geom

    It replaces `{usage}`, `{common_parameters}` and
    `{aesthetics}` with generated documentation.
    """
    # Dedented so that it lineups (in sphinx) with the part
    # generated parts when put together
    docstring = dedent(geom.__doc__ or "")

    # usage
    signature = make_signature(
        geom.__name__,
        geom.DEFAULT_PARAMS,
        common_geom_params,
        common_geom_param_values,
    )
    usage = GEOM_SIGNATURE_TPL.format(signature=signature)

    # aesthetics
    contents = {f"**{ae}**": "" for ae in sorted(geom.REQUIRED_AES)}
    if geom.DEFAULT_AES:
        d = geom.DEFAULT_AES.copy()
        d["group"] = ""  # All geoms understand the group aesthetic
        contents.update(sorted(d.items()))

    table = dict_to_table(("Aesthetic", "Default value"), contents)
    aesthetics_table = AESTHETICS_TABLE_TPL.format(table=table)
    tpl = dedent(geom._aesthetics_doc.lstrip("\n"))
    aesthetics_doc = tpl.format(aesthetics_table=aesthetics_table)
    aesthetics_doc = indent(aesthetics_doc, " " * 4)

    # common_parameters
    d = geom.DEFAULT_PARAMS
    common_parameters = GEOM_PARAMS_TPL.format(
        default_stat=qualified_name(d["stat"], "stat_"),
        default_position=qualified_name(d["position"], "position_"),
        default_na_rm=d["na_rm"],
        default_inherit_aes=d.get("inherit_aes", True),
        default_raster=d.get("raster", False),
        _aesthetics_doc=aesthetics_doc,
        **common_params_doc,
    )

    docstring = docstring.replace("{usage}", usage)
    docstring = docstring.replace("{common_parameters}", common_parameters)
    geom.__doc__ = docstring
    return geom


def document_stat(stat: type[Stat]) -> type[Stat]:
    """
    Create a structured documentation for the stat

    It replaces `{usage}`, `{common_parameters}` and
    `{aesthetics}` with generated documentation.
    """
    # Dedented so that it lineups (in sphinx) with the part
    # generated parts when put together
    docstring = dedent(stat.__doc__ or "")

    # usage:
    signature = make_signature(
        stat.__name__,
        stat.DEFAULT_PARAMS,
        common_stat_params,
        common_stat_param_values,
    )
    usage = STAT_SIGNATURE_TPL.format(signature=signature)

    # aesthetics
    contents = {f"**{ae}**": "" for ae in sorted(stat.REQUIRED_AES)}
    contents.update(sorted(stat.DEFAULT_AES.items()))
    table = dict_to_table(("Aesthetic", "Default value"), contents)
    aesthetics_table = AESTHETICS_TABLE_TPL.format(table=table)
    tpl = dedent(stat._aesthetics_doc.lstrip("\n"))
    aesthetics_doc = tpl.format(aesthetics_table=aesthetics_table)
    aesthetics_doc = indent(aesthetics_doc, " " * 4)

    # common_parameters
    d = stat.DEFAULT_PARAMS
    common_parameters = STAT_PARAMS_TPL.format(
        default_geom=qualified_name(d["geom"], "geom_"),
        default_position=qualified_name(d["position"], "position_"),
        default_na_rm=d["na_rm"],
        _aesthetics_doc=aesthetics_doc,
        **common_params_doc,
    )

    docstring = docstring.replace("{usage}", usage)
    docstring = docstring.replace("{common_parameters}", common_parameters)
    stat.__doc__ = docstring
    return stat


def document_scale(cls: type[Scale]) -> type[Scale]:
    """
    Create a documentation for a scale

    Import the superclass parameters

    It replaces `{superclass_parameters}` with the documentation
    of the parameters from the superclass.

    Parameters
    ----------
    cls : type
        A scale class

    Returns
    -------
    cls : type
        The scale class with a modified docstring.
    """
    params_list = []
    # Get set of cls params
    cls_param_string = docstring_parameters_section(cls)
    cls_param_dict = parameters_str_to_dict(cls_param_string)
    cls_params = set(cls_param_dict.keys())

    for i, base in enumerate(cls.__bases__):
        # Get set of base class params
        base_param_string = param_string = docstring_parameters_section(base)
        base_param_dict = parameters_str_to_dict(base_param_string)
        base_params = set(base_param_dict.keys())

        # Remove duplicate params from the base class
        duplicate_params = base_params & cls_params
        for param in duplicate_params:
            del base_param_dict[param]

        if duplicate_params:
            param_string = parameters_dict_to_str(base_param_dict)

        # Accumulate params of base case
        if i == 0:
            # Compensate for the indentation of the
            # {superclass_parameters} string
            param_string = param_string.strip()
        params_list.append(param_string)

        # Prevent the next base classes from bringing in the
        # same parameters.
        cls_params |= base_params

    # Fill in the processed superclass parameters
    superclass_parameters = "\n".join(params_list)
    cls_doc = cls.__doc__ or ""  # for typechecker
    cls.__doc__ = cls_doc.format(superclass_parameters=superclass_parameters)
    return cls


DOC_FUNCTIONS = {
    "geom": document_geom,
    "stat": document_stat,
    "scale": document_scale,
}


def document(cls: Type[T]) -> Type[T]:
    """
    Document a plotnine class

    To be used as a decorator
    """
    if cls.__doc__ is None:
        return cls

    baseclass_name = cls.mro()[-2].__name__

    try:
        return DOC_FUNCTIONS[baseclass_name](cls)
    except KeyError:
        return cls



from __future__ import annotations

import typing
from warnings import warn

import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineWarning
from ..mapping import aes
from ..utils import order_as_data_mapping
from .geom import geom
from .geom_path import geom_path
from .geom_segment import geom_segment

if typing.TYPE_CHECKING:
    from typing import Any

    from plotnine.iapi import panel_view
    from plotnine.typing import (
        Aes,
        Axes,
        Coord,
        DataLike,
    )


@document
class geom_hline(geom):
    """
    Horizontal line

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_AES = {
        "color": "black",
        "linetype": "solid",
        "size": 0.5,
        "alpha": 1,
    }
    REQUIRED_AES = {"yintercept"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "inherit_aes": False,
    }
    draw_legend = staticmethod(geom_path.draw_legend)

    def __init__(
        self,
        mapping: Aes | None = None,
        data: DataLike | None = None,
        **kwargs: Any,
    ):
        data, mapping = order_as_data_mapping(data, mapping)
        yintercept = kwargs.pop("yintercept", None)
        if yintercept is not None:
            if mapping:
                warn(
                    "The 'yintercept' parameter has overridden "
                    "the aes() mapping.",
                    PlotnineWarning,
                )
            data = pd.DataFrame({"yintercept": np.repeat(yintercept, 1)})
            mapping = aes(yintercept="yintercept")
            kwargs["show_legend"] = False

        geom.__init__(self, mapping, data, **kwargs)

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        """
        Plot all groups
        """
        ranges = coord.backtransform_range(panel_params)
        data["y"] = data["yintercept"]
        data["yend"] = data["yintercept"]
        data["x"] = ranges.x[0]
        data["xend"] = ranges.x[1]
        data = data.drop_duplicates()

        for _, gdata in data.groupby("group"):
            gdata.reset_index(inplace=True)
            geom_segment.draw_group(gdata, panel_params, coord, ax, **params)



from __future__ import annotations

import typing

from ..doctools import document
from ..utils import resolution
from .geom_rect import geom_rect

if typing.TYPE_CHECKING:
    import pandas as pd


@document
class geom_tile(geom_rect):
    """
    Rectangles specified using a center points

    {usage}

    Parameters
    ----------
    {common_parameters}

    See Also
    --------
    plotnine.geoms.geom_rect
    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": None,
        "fill": "#333333",
        "linetype": "solid",
        "size": 0.1,
    }
    REQUIRED_AES = {"x", "y"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
    }

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        try:
            width = data.pop("width")
        except KeyError:
            width = resolution(data["x"], False)

        try:
            height = data.pop("height")
        except KeyError:
            height = resolution(data["y"], False)

        data["xmin"] = data["x"] - width / 2
        data["xmax"] = data["x"] + width / 2
        data["ymin"] = data["y"] - height / 2
        data["ymax"] = data["y"] + height / 2
        return data



from __future__ import annotations

import typing

import pandas as pd

from ..exceptions import PlotnineError
from ..utils import add_margins, cross_join, join_keys, match, ninteraction
from .facet import (
    add_missing_facets,
    combine_vars,
    eval_facet_vars,
    facet,
    layout_null,
)
from .strips import Strips, strip

if typing.TYPE_CHECKING:
    from typing import Literal

    from plotnine.iapi import layout_details
    from plotnine.typing import Axes


class facet_grid(facet):
    """
    Wrap 1D Panels onto 2D surface

    Parameters
    ----------
    facets : str | tuple | list
        A formula with the rows (of the tabular display) on
        the LHS and the columns (of the tabular display) on
        the RHS; the dot in the formula is used to indicate
        there should be no faceting on this dimension
        (either row or column). If a tuple/list is used, it
        must of size two, the elements of which must be
        strings or lists. If string formula is not processed
        as you may expect, use tuple/list. For example, the
        follow two specifications are equivalent::

            'func(var4) ~ func(var1+var3) + func(var2)'
            ['func(var4)', ('func(var1+var3)', 'func(var2)')]

        There may be cases where you cannot use a
        use a pure string formula, e.g.::

            ['var4', ('var1+var3', 'var2')]

    margins : bool | list[str]
        variable names to compute margins for.
        True will compute all possible margins.
    space : str | dict
        Control the size of the  ``x`` or ``y`` sides of the panels.
        The size also depends to the ``scales`` parameter.

        If a string, it should be one of
        ``['fixed', 'free', 'free_x', 'free_y']``. Currently, only the
        ``'fixed'`` option is supported.

        Alternatively if a ``dict``, it indicates the relative facet
        size ratios such as::

            {'x': [1, 2], 'y': [3, 1, 1]}

        This means that in the horizontal direction, the second panel
        will be twice the length of the first. In the vertical direction
        the top facet will be the 3 times longer then the second and
        third facets.

        Note that the number of dimensions in the list must equal the
        number of facets that will be produced.

    shrink : bool
        Whether to shrink the scales to the output of the
        statistics instead of the raw data. Default is ``True``.
    labeller : str | function
        How to label the facets. If it is a ``str``, it should
        be one of ``'label_value'`` ``'label_both'`` or
        ``'label_context'``. Default is ``'label_value'``
    as_table : bool
        If ``True``, the facets are laid out like a table with
        the highest values at the bottom-right. If ``False``
        the facets are laid out like a plot with the highest
        value a the top-right. Default it ``True``.
    drop : bool
        If ``True``, all factor levels not used in the data
        will automatically be dropped. If ``False``, all
        factor levels will be shown, regardless of whether
        or not they appear in the data. Default is ``True``.
    """

    rows: list[str]
    cols: list[str]

    def __init__(
        self,
        facets: str | tuple[str | list[str], str | list[str]],
        margins: bool | list[str] = False,
        scales: Literal["fixed", "free", "free_x", "free_y"] = "fixed",
        space: (
            Literal["fixed", "free", "free_x", "free_y"]
            | dict[Literal["x", "y"], list[int]]
        ) = "fixed",
        shrink: bool = True,
        labeller: Literal[
            "label_value", "label_both", "label_context"
        ] = "label_value",
        as_table: bool = True,
        drop: bool = True,
    ):
        facet.__init__(
            self,
            scales=scales,
            shrink=shrink,
            labeller=labeller,
            as_table=as_table,
            drop=drop,
        )
        self.space = space
        self.rows, self.cols = parse_grid_facets(facets)
        self.margins = margins
        self.space_free = {
            "x": isinstance(space, str) and space in ("free_x", "free"),
            "y": isinstance(space, str) and space in ("free_y", "free"),
        }
        self.num_vars_x = len(self.cols)
        self.num_vars_y = len(self.rows)

    def compute_layout(self, data: list[pd.DataFrame]) -> pd.DataFrame:
        if not self.rows and not self.cols:
            return layout_null()

        base_rows = combine_vars(
            data, self.plot.environment, self.rows, drop=self.drop
        )

        if not self.as_table:
            # Reverse the order of the rows
            base_rows = base_rows[::-1]
        base_cols = combine_vars(
            data, self.plot.environment, self.cols, drop=self.drop
        )

        base = cross_join(base_rows, base_cols)

        if self.margins:
            base = add_margins(base, (self.rows, self.cols), self.margins)
            base = base.drop_duplicates().reset_index(drop=True)

        n = len(base)
        panel = ninteraction(base, drop=True)
        panel = pd.Categorical(panel, categories=range(1, n + 1))

        if self.rows:
            rows = ninteraction(base[self.rows], drop=True)
        else:
            rows = [1] * len(panel)

        if self.cols:
            cols = ninteraction(base[self.cols], drop=True)
        else:
            cols = [1] * len(panel)

        layout = pd.DataFrame(
            {
                "PANEL": panel,
                "ROW": rows,
                "COL": cols,
            }
        )
        layout = pd.concat([layout, base], axis=1)
        layout = layout.sort_values("PANEL")
        layout.reset_index(drop=True, inplace=True)

        # Relax constraints, if necessary
        layout["SCALE_X"] = layout["COL"] if self.free["x"] else 1
        layout["SCALE_Y"] = layout["ROW"] if self.free["y"] else 1
        layout["AXIS_X"] = layout["ROW"] == layout["ROW"].max()
        layout["AXIS_Y"] = layout["COL"] == layout["COL"].min()

        self.nrow = layout["ROW"].max()
        self.ncol = layout["COL"].max()
        return layout

    def map(self, data: pd.DataFrame, layout: pd.DataFrame) -> pd.DataFrame:
        if not len(data):
            data["PANEL"] = pd.Categorical(
                [], categories=layout["PANEL"].cat.categories, ordered=True
            )
            return data

        vars = list(self.rows + self.cols)
        margin_vars: tuple[list[str], list[str]] = (
            list(data.columns.intersection(self.rows)),
            list(data.columns.intersection(self.cols)),
        )
        data = add_margins(data, margin_vars, self.margins)

        facet_vals = eval_facet_vars(data, vars, self.plot.environment)
        data, facet_vals = add_missing_facets(data, layout, vars, facet_vals)

        # assign each point to a panel
        if len(facet_vals) == 0:
            # Special case of no facetting
            data["PANEL"] = 1
        else:
            keys = join_keys(facet_vals, layout, vars)
            data["PANEL"] = match(keys["x"], keys["y"], start=1)

        # matching dtype and
        # the categories(panel numbers) for the data should be in the
        # same order as the panels. i.e the panels are the reference,
        # they "know" the right order
        data["PANEL"] = pd.Categorical(
            data["PANEL"],
            categories=layout["PANEL"].cat.categories,
            ordered=True,
        )

        data.reset_index(drop=True, inplace=True)
        return data

    def make_ax_strips(self, layout_info: layout_details, ax: Axes) -> Strips:
        lst = []
        toprow = layout_info.row == 1
        rightcol = layout_info.col == self.ncol
        if toprow and len(self.cols):
            s = strip(self.cols, layout_info, self, ax, "top")
            lst.append(s)
        if rightcol and len(self.rows):
            s = strip(self.rows, layout_info, self, ax, "right")
            lst.append(s)
        return Strips(lst)


def parse_grid_facets(
    facets: str | tuple[str | list[str], str | list[str]]
) -> tuple[list[str], list[str]]:
    """
    Return two lists of facetting variables, for the rows & columns
    """
    valid_seqs = [
        "('var1', '.')",
        "('var1', 'var2')",
        "('.', 'var1')",
        "((var1, var2), (var3, var4))",
    ]
    error_msg_s = (
        "Valid sequences for specifying 'facets' look like" f" {valid_seqs}"
    )

    valid_forms = [
        "var1 ~ .",
        "var1 ~ var2",
        ". ~ var1",
        "var1 + var2 ~ var3 + var4",
        ". ~ func(var1) + func(var2)",
        ". ~ func(var1+var3) + func(var2)",
    ] + valid_seqs
    error_msg_f = "Valid formula for 'facet_grid' look like" f" {valid_forms}"

    if not isinstance(facets, str):
        if len(facets) != 2:
            raise PlotnineError(error_msg_s)

        rows = ensure_list_spec(facets[0])
        cols = ensure_list_spec(facets[1])

        return rows, cols

    # Example of allowed formulae
    # "c ~ a + b'
    # '. ~ func(a) + func(b)'
    # 'func(c) ~ func(a+1) + func(b+2)'
    try:
        lhs, rhs = facets.split("~")
    except ValueError:
        raise PlotnineError(error_msg_f)
    else:
        lhs = lhs.strip()
        rhs = rhs.strip()

    rows = ensure_list_spec(lhs)
    cols = ensure_list_spec(rhs)
    return rows, cols


def ensure_list_spec(term: list[str] | str) -> list[str]:
    """
    Convert a str specification to a list spec

    e.g.
    'a' -> ['a']
    'a + b' -> ['a', 'b']
    '.' -> []
    '' -> []
    """
    if isinstance(term, str):
        splitter = " + " if " + " in term else "+"
        if term in [".", ""]:
            return []
        return [var.strip() for var in term.split(splitter)]
    else:
        return term



import numpy as np
import pandas as pd
import pytest

from plotnine import (
    aes,
    arrow,
    facet_grid,
    geom_line,
    geom_path,
    geom_point,
    geom_step,
    ggplot,
)
from plotnine.exceptions import PlotnineWarning

# steps with diagonals at the ends
df = pd.DataFrame(
    {
        "x": [1, 2, 3, 3, 4, 4, 5, 5, 6, 7],
        "y": [1, 2, 2, 3, 3, 4, 4, 5, 5, 6],
    }
)


def test_aesthetics():
    p = (
        ggplot(df, aes("x", "y"))
        + geom_path(size=4)
        + geom_path(aes(y="y+2", alpha="x"), size=4, show_legend=False)
        + geom_path(aes(y="y+4"), size=4, linetype="dashed", show_legend=False)
        + geom_path(aes(y="y+6", size="x"), color="red", show_legend=False)
        + geom_path(aes(y="y+8", color="x"), size=4)
    )

    assert p == "aesthetics"


def test_arrow():
    p = (
        ggplot(df, aes("x", "y"))
        + geom_path(size=2, arrow=arrow(ends="both", type="closed"))
        + geom_path(
            aes(y="y+2"),
            color="red",
            size=2,
            arrow=arrow(angle=60, length=0.8, ends="first"),
        )
        + geom_path(
            aes(y="y+4"), color="blue", size=2, arrow=arrow(length=0.8)
        )
    )

    assert p == "arrow"


def test_arrow_facets():
    df = pd.DataFrame(
        {"x": [1, 3, 2, 4], "y": [10, 9, 10, 9], "z": ["a", "a", "b", "b"]}
    )

    p = (
        ggplot(df, aes("x", "y"))
        + geom_path(size=2, arrow=arrow(length=0.25))
        + facet_grid("~ z")
    )
    assert p == "arrow_facets"


def test_step():
    p = (
        ggplot(df, aes("x"))
        + geom_step(aes(y="y"), size=4)
        + geom_step(aes(y="y+2"), color="red", direction="vh", size=4)
    )

    assert p == "step"


def test_step_mid():
    df = pd.DataFrame({"x": range(9), "y": range(9)})
    p = (
        ggplot(df, aes("x", "y"))
        + geom_point(size=4)
        + geom_step(direction="mid", size=2)
    )

    assert p == "step_mid"


def test_line():
    df2 = df.copy()

    # geom_path plots in given order. geom_line &
    # geom_step sort by x before plotting
    df2["x"] = df["x"].values[::-1]

    p = (
        ggplot(df2, aes("x"))
        + geom_path(aes(y="y"), size=4)
        + geom_line(aes(y="y+2"), color="blue", size=4)
        + geom_step(aes(y="y+4"), color="red", size=4)
    )

    assert p == "path_line_step"


df_missing = pd.DataFrame(
    {
        "x": [1, 2, 3, 4, 5, 6, 7],
        "y1": [np.nan, 1, 2, 3, 4, np.nan, np.nan],
        "y2": [1, 2, 3, np.nan, np.nan, 6, 7],
    }
)


def test_missing_values():
    p = ggplot(df_missing, aes(x="x")) + geom_line(aes(y="y1"), size=2)

    with pytest.warns(PlotnineWarning):
        assert p == "missing_values"


def test_no_missing_values():
    p = ggplot(df_missing, aes(x="x")) + geom_line(aes(y="y2"), size=2)

    assert p == "no_missing_values"


def test_groups_less_that_two_points():
    df = pd.DataFrame(
        {
            "A": [1, 2, 3, 4, 5],
            "B": [0, 0, 1, 2, 2],
            "C": [1, 2, 3, 4, 5],
            "D": [1, 2, 3, 4, 5],
        }
    )

    p = ggplot(df) + geom_line(aes(x="A", y="C", group="B", color="D"), size=2)
    p.draw_test()



from __future__ import annotations

import re
import typing
from warnings import warn

import numpy as np
import pandas as pd

from ..exceptions import PlotnineError, PlotnineWarning
from ..utils import join_keys, match
from .facet import (
    add_missing_facets,
    combine_vars,
    eval_facet_vars,
    facet,
    layout_null,
)
from .strips import Strips, strip

if typing.TYPE_CHECKING:
    from typing import Literal, Optional

    from plotnine.iapi import layout_details
    from plotnine.typing import Axes


class facet_wrap(facet):
    """
    Wrap 1D Panels onto 2D surface

    Parameters
    ----------
    facets : formula | tuple | list
        Variables to groupby and plot on different panels.
        If a formula is used it should be right sided,
        e.g ``'~ a + b'``, ``('a', 'b')``
    nrow : int, optional
        Number of rows
    ncol : int, optional
        Number of columns
    scales : str in ``['fixed', 'free', 'free_x', 'free_y']``
        Whether ``x`` or ``y`` scales should be allowed (free)
        to vary according to the data on each of the panel.
        Default is ``'fixed'``.
    shrink : bool
        Whether to shrink the scales to the output of the
        statistics instead of the raw data. Default is ``True``.
    labeller : str | function
        How to label the facets. If it is a ``str``, it should
        be one of ``'label_value'`` ``'label_both'`` or
        ``'label_context'``. Default is ``'label_value'``
    as_table : bool
        If ``True``, the facets are laid out like a table with
        the highest values at the bottom-right. If ``False``
        the facets are laid out like a plot with the highest
        value a the top-right. Default it ``True``.
    drop : bool
        If ``True``, all factor levels not used in the data
        will automatically be dropped. If ``False``, all
        factor levels will be shown, regardless of whether
        or not they appear in the data. Default is ``True``.
    dir : str in ``['h', 'v']``
        Direction in which to layout the panels. ``h`` for
        horizontal and ``v`` for vertical.
    """

    def __init__(
        self,
        facets: str | list[str],
        *,
        nrow: Optional[int] = None,
        ncol: Optional[int] = None,
        scales: Literal["fixed", "free", "free_x", "free_y"] = "fixed",
        shrink: bool = True,
        labeller: Literal[
            "label_value", "label_both", "label_context"
        ] = "label_value",
        as_table: bool = True,
        drop: bool = True,
        dir: Literal["h", "v"] = "h",
    ):
        super().__init__(
            scales=scales,
            shrink=shrink,
            labeller=labeller,
            as_table=as_table,
            drop=drop,
            dir=dir,
        )
        self.vars = parse_wrap_facets(facets)
        self._nrow, self._ncol = check_dimensions(nrow, ncol)
        # facet_wrap gets its labelling at the top
        self.num_vars_x = len(self.vars)

    def compute_layout(
        self,
        data: list[pd.DataFrame],
    ) -> pd.DataFrame:
        if not self.vars:
            return layout_null()

        base = combine_vars(
            data, self.plot.environment, self.vars, drop=self.drop
        )
        n = len(base)
        dims = wrap_dims(n, self._nrow, self._ncol)
        _id = np.arange(1, n + 1)

        if self.dir == "v":
            dims = dims[::-1]

        if self.as_table:
            row = (_id - 1) // dims[1] + 1
        else:
            row = dims[0] - (_id - 1) // dims[1]

        col = (_id - 1) % dims[1] + 1

        layout = pd.DataFrame(
            {
                "PANEL": pd.Categorical(range(1, n + 1)),
                "ROW": row.astype(int),
                "COL": col.astype(int),
            }
        )
        if self.dir == "v":
            layout.rename(columns={"ROW": "COL", "COL": "ROW"}, inplace=True)

        layout = pd.concat([layout, base], axis=1)
        self.nrow = layout["ROW"].nunique()
        self.ncol = layout["COL"].nunique()
        n = layout.shape[0]

        # Add scale identification
        layout["SCALE_X"] = range(1, n + 1) if self.free["x"] else 1
        layout["SCALE_Y"] = range(1, n + 1) if self.free["y"] else 1

        # Figure out where axes should go.
        # The bottom-most row of each column and the left most
        # column of each row
        x_idx = [df["ROW"].idxmax() for _, df in layout.groupby("COL")]
        y_idx = [df["COL"].idxmin() for _, df in layout.groupby("ROW")]
        layout["AXIS_X"] = False
        layout["AXIS_Y"] = False
        _loc = layout.columns.get_loc
        layout.iloc[x_idx, _loc("AXIS_X")] = True  # type: ignore
        layout.iloc[y_idx, _loc("AXIS_Y")] = True  # type: ignore

        if self.free["x"]:
            layout.loc[:, "AXIS_X"] = True

        if self.free["y"]:
            layout.loc[:, "AXIS_Y"] = True

        return layout

    def map(self, data: pd.DataFrame, layout: pd.DataFrame) -> pd.DataFrame:
        if not len(data):
            data["PANEL"] = pd.Categorical(
                [], categories=layout["PANEL"].cat.categories, ordered=True
            )
            return data

        facet_vals = eval_facet_vars(data, self.vars, self.plot.environment)
        data, facet_vals = add_missing_facets(
            data, layout, self.vars, facet_vals
        )

        # assign each point to a panel
        keys = join_keys(facet_vals, layout, self.vars)
        data["PANEL"] = match(keys["x"], keys["y"], start=1)

        # matching dtype
        data["PANEL"] = pd.Categorical(
            data["PANEL"],
            categories=layout["PANEL"].cat.categories,
            ordered=True,
        )

        data.reset_index(drop=True, inplace=True)
        return data

    def make_ax_strips(self, layout_info: layout_details, ax: Axes) -> Strips:
        s = strip(self.vars, layout_info, self, ax, "top")
        return Strips([s])


def check_dimensions(
    nrow: Optional[int], ncol: Optional[int]
) -> tuple[int | None, int | None]:
    """
    Verify dimensions of the facet
    """
    if nrow is not None:
        if nrow < 1:
            warn(
                "'nrow' must be greater than 0. "
                "Your value has been ignored.",
                PlotnineWarning,
            )
            nrow = None
        else:
            nrow = int(nrow)

    if ncol is not None:
        if ncol < 1:
            warn(
                "'ncol' must be greater than 0. "
                "Your value has been ignored.",
                PlotnineWarning,
            )
            ncol = None
        else:
            ncol = int(ncol)

    return nrow, ncol


def parse_wrap_facets(facets: str | list[str]) -> list[str]:
    """
    Return list of facetting variables
    """
    valid_forms = ["~ var1", "~ var1 + var2"]
    error_msg = "Valid formula for 'facet_wrap' look like" f" {valid_forms}"

    if isinstance(facets, (list, tuple)):
        return facets

    if not isinstance(facets, str):
        raise PlotnineError(error_msg)

    if "~" in facets:
        variables_pattern = r"(\w+(?:\s*\+\s*\w+)*|\.)"
        pattern = rf"\s*~\s*{variables_pattern}\s*"
        match = re.match(pattern, facets)
        if not match:
            raise PlotnineError(error_msg)

        facets = [var.strip() for var in match.group(1).split("+")]
    elif re.match(r"\w+", facets):
        # allow plain string as the variable name
        facets = [facets]
    else:
        raise PlotnineError(error_msg)

    return facets


def wrap_dims(
    n: int, nrow: Optional[int] = None, ncol: Optional[int] = None
) -> tuple[int, int]:
    """
    Wrap dimensions
    """
    if nrow is None:
        if ncol is None:
            return n_to_nrow_ncol(n)
        else:
            nrow = int(np.ceil(n / ncol))

    if ncol is None:
        ncol = int(np.ceil(n / nrow))

    if not nrow * ncol >= n:
        raise PlotnineError(
            "Allocated fewer panels than are required. "
            "Make sure the number of rows and columns can "
            "hold all the plot panels."
        )
    return (nrow, ncol)


def n_to_nrow_ncol(n: int) -> tuple[int, int]:
    """
    Compute the rows and columns given the number of plots.
    """
    if n <= 3:
        nrow, ncol = 1, n
    elif n <= 6:
        nrow, ncol = 2, (n + 1) // 2
    elif n <= 12:
        nrow, ncol = 3, (n + 2) // 3
    else:
        ncol = int(np.ceil(np.sqrt(n)))
        nrow = int(np.ceil(n / ncol))
    return (nrow, ncol)



from __future__ import annotations

import typing
from warnings import warn

import numpy as np

from ..doctools import document
from ..exceptions import PlotnineWarning
from ..utils import groupby_apply, resolution, to_rgba
from .geom import geom

if typing.TYPE_CHECKING:
    from typing import Any

    import pandas as pd

    from plotnine.iapi import panel_view
    from plotnine.typing import Axes, Coord, DrawingArea, Layer


@document
class geom_dotplot(geom):
    """
    Dot plot

    {usage}

    Parameters
    ----------
    {common_parameters}
    stackdir : str (default: up)
        Direction in which to stack the dots. Options are
        :py:`['up', 'down', 'center', 'centerwhole']`
    stackratio : float (default: 1)
        How close to stack the dots. If value is less than 1,
        the dots overlap, if greater than 1 they are spaced.
    dotsize : float (default: 1)
        Diameter of dots relative to ``binwidth``.
    stackgroups : bool (default: False)
        If :py:`True`, the dots are stacked across groups.

    See Also
    --------
    plotnine.stats.stat_bindot
    """

    DEFAULT_AES = {"alpha": 1, "color": "black", "fill": "black"}
    REQUIRED_AES = {"x", "y"}
    NON_MISSING_AES = {"size", "shape"}
    DEFAULT_PARAMS = {
        "stat": "bindot",
        "position": "identity",
        "na_rm": False,
        "stackdir": "up",
        "stackratio": 1,
        "dotsize": 1,
        "stackgroups": False,
    }

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        gp = self.params
        sp = self._stat.params

        # Issue warnings when parameters don't make sense
        if gp["position"] == "stack":
            warn(
                "position='stack' doesn't work properly with "
                "geom_dotplot. Use stackgroups=True instead.",
                PlotnineWarning,
            )
        if (
            gp["stackgroups"]
            and sp["method"] == "dotdensity"
            and sp["binpositions"] == "bygroup"
        ):
            warn(
                "geom_dotplot called with stackgroups=TRUE and "
                "method='dotdensity'. You probably want to set "
                "binpositions='all'",
                PlotnineWarning,
            )

        if "width" not in data:
            if sp["width"]:
                data["width"] = sp["width"]
            else:
                data["width"] = resolution(data["x"], False) * 0.9

        # Set up the stacking function and range
        if gp["stackdir"] in (None, "up"):

            def stackdots(a: pd.Series[float]) -> pd.Series[float]:
                return a - 0.5

            stackaxismin: float = 0
            stackaxismax: float = 1
        elif gp["stackdir"] == "down":

            def stackdots(a: pd.Series[float]) -> pd.Series[float]:
                return -a + 0.5

            stackaxismin = -1
            stackaxismax = 0
        elif gp["stackdir"] == "center":

            def stackdots(a: pd.Series[float]) -> pd.Series[float]:
                return a - 1 - np.max(a - 1) / 2

            stackaxismin = -0.5
            stackaxismax = 0.5
        elif gp["stackdir"] == "centerwhole":

            def stackdots(a: pd.Series[float]) -> pd.Series[float]:
                return a - 1 - np.floor(np.max(a - 1) / 2)

            stackaxismin = -0.5
            stackaxismax = 0.5
        else:
            raise ValueError(f"Invalid value stackdir={gp['stackdir']}")

        # Fill the bins: at a given x (or y),
        # if count=3, make 3 entries at that x
        idx = [i for i, c in enumerate(data["count"]) for j in range(int(c))]
        data = data.iloc[idx]
        data.reset_index(inplace=True, drop=True)
        # Next part will set the position of each dot within each stack
        # If stackgroups=TRUE, split only on x (or y) and panel;
        # if not stacking, also split by group
        groupvars = [sp["binaxis"], "PANEL"]
        if not gp["stackgroups"]:
            groupvars.append("group")

        # Within each x, or x+group, set countidx=1,2,3,
        # and set stackpos according to stack function
        def func(df: pd.DataFrame) -> pd.DataFrame:
            df["countidx"] = range(1, len(df) + 1)
            df["stackpos"] = stackdots(df["countidx"])
            return df

        # Within each x, or x+group, set countidx=1,2,3, and set
        # stackpos according to stack function
        data = groupby_apply(data, groupvars, func)

        # Set the bounding boxes for the dots
        if sp["binaxis"] == "x":
            # ymin, ymax, xmin, and xmax define the bounding
            # rectangle for each stack. Can't do bounding box per dot,
            # because y position isn't real.
            # After position code is rewritten, each dot should have
            # its own bounding box.
            data["xmin"] = data["x"] - data["binwidth"] / 2
            data["xmax"] = data["x"] + data["binwidth"] / 2
            data["ymin"] = stackaxismin
            data["ymax"] = stackaxismax
            data["y"] = 0
        elif sp["binaxis"] == "y":
            # ymin, ymax, xmin, and xmax define the bounding
            # rectangle for each stack. Can't do bounding box per dot,
            # because x position isn't real.
            # xmin and xmax aren't really the x bounds. They're just
            # set to the standard x +- width/2 so that dot clusters
            # can be dodged like other geoms.
            # After position code is rewritten, each dot should have
            # its own bounding box.
            def func(df: pd.DataFrame) -> pd.DataFrame:
                df["ymin"] = df["y"].min() - data["binwidth"][0] / 2
                df["ymax"] = df["y"].max() + data["binwidth"][0] / 2
                return df

            data = groupby_apply(data, "group", func)
            data["xmin"] = data["x"] + data["width"] * stackaxismin
            data["xmax"] = data["x"] + data["width"] * stackaxismax

        return data

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        from matplotlib.collections import PatchCollection
        from matplotlib.patches import Ellipse

        data = coord.transform(data, panel_params)
        fill = to_rgba(data["fill"], data["alpha"])
        color = to_rgba(data["color"], data["alpha"])
        ranges = coord.range(panel_params)

        # For perfect circles the width/height of the circle(ellipse)
        # should factor in the dimensions of axes
        bbox = ax.get_window_extent().transformed(
            ax.figure.dpi_scale_trans.inverted()  # pyright: ignore
        )
        ax_width, ax_height = bbox.width, bbox.height

        factor = (ax_width / ax_height) * np.ptp(ranges.y) / np.ptp(ranges.x)
        size = data.loc[0, "binwidth"] * params["dotsize"]
        offsets = data["stackpos"] * params["stackratio"]

        if params["binaxis"] == "x":
            width, height = size, size * factor
            xpos, ypos = data["x"], data["y"] + height * offsets
        elif params["binaxis"] == "y":
            width, height = size / factor, size
            xpos, ypos = data["x"] + width * offsets, data["y"]
        else:
            raise ValueError(
                f"Invalid valid value binaxis={params['binaxis']}"
            )

        circles = []
        for xy in zip(xpos, ypos):
            patch = Ellipse(xy, width=width, height=height)
            circles.append(patch)

        coll = PatchCollection(
            circles,
            edgecolors=color,
            facecolors=fill,
            rasterized=params["raster"],
        )
        ax.add_collection(coll)

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: Layer
    ) -> DrawingArea:
        """
        Draw a point in the box

        Parameters
        ----------
        data : Series
            Data Row
        da : DrawingArea
            Canvas
        lyr : layer
            Layer

        Returns
        -------
        out : DrawingArea
        """
        from matplotlib.lines import Line2D

        fill = to_rgba(data["fill"], data["alpha"])
        key = Line2D(
            [0.5 * da.width],
            [0.5 * da.height],
            marker="o",
            markersize=da.width / 2,
            markerfacecolor=fill,
            markeredgecolor=data["color"],
        )
        da.add_artist(key)
        return da



from __future__ import annotations

import typing

import numpy as np
import pandas as pd

from ..doctools import document
from ..utils import groupby_apply, interleave, resolution
from .geom import geom
from .geom_path import geom_path
from .geom_polygon import geom_polygon

if typing.TYPE_CHECKING:
    from typing import Any

    from plotnine.iapi import panel_view
    from plotnine.typing import Aes, Axes, Coord, DataLike, FloatArray


@document
class geom_violin(geom):
    """
    Violin Plot

    {usage}

    Parameters
    ----------
    {common_parameters}
    draw_quantiles : float or [float]
       draw horizontal lines at the given quantiles (0..1)
       of the density estimate.
    style : str, optional (default: 'full')
       The type of violin plot to draw. The options are:

       ::

           'full'        # Regular (2 sided violins)
           'left'        # Left-sided half violins
           'right'       # Right-sided half violins
           'left-right'  # Alternate (left first) half violins by the group
           'right-left'  # Alternate (right first) half violins by the group

    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": "#333333",
        "fill": "white",
        "linetype": "solid",
        "size": 0.5,
        "weight": 1,
    }
    REQUIRED_AES = {"x", "y"}
    DEFAULT_PARAMS = {
        "stat": "ydensity",
        "position": "dodge",
        "draw_quantiles": None,
        "style": "full",
        "scale": "area",
        "trim": True,
        "width": None,
        "na_rm": False,
    }
    draw_legend = staticmethod(geom_polygon.draw_legend)

    def __init__(
        self,
        mapping: Aes | None = None,
        data: DataLike | None = None,
        **kwargs: Any,
    ):
        if "draw_quantiles" in kwargs:
            kwargs["draw_quantiles"] = np.repeat(kwargs["draw_quantiles"], 1)
            if not all(0 < q < 1 for q in kwargs["draw_quantiles"]):
                raise ValueError(
                    "draw_quantiles must be a float or "
                    "an iterable of floats (>0.0; < 1.0)"
                )

        if "style" in kwargs:
            allowed = ("full", "left", "right", "left-right", "right-left")
            if kwargs["style"] not in allowed:
                raise ValueError(f"style must be either {allowed}")

        super().__init__(mapping, data, **kwargs)

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        if "width" not in data:
            if self.params["width"]:
                data["width"] = self.params["width"]
            else:
                data["width"] = resolution(data["x"], False) * 0.9

        def func(df: pd.DataFrame) -> pd.DataFrame:
            df["ymin"] = df["y"].min()
            df["ymax"] = df["y"].max()
            df["xmin"] = df["x"] - df["width"] / 2
            df["xmax"] = df["x"] + df["width"] / 2
            return df

        # This is a plyr::ddply
        data = groupby_apply(data, ["group", "PANEL"], func)
        return data

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        quantiles = params["draw_quantiles"]
        style = params["style"]

        for i, (_, df) in enumerate(data.groupby("group")):
            # Find the points for the line to go all the way around
            df["xminv"] = df["x"] - df["violinwidth"] * (df["x"] - df["xmin"])
            df["xmaxv"] = df["x"] + df["violinwidth"] * (df["xmax"] - df["x"])
            even = i % 2 == 0
            if (
                style == "left"
                or (style == "left-right" and even)
                or (style == "right-left" and not even)
            ):
                df["xmaxv"] = df["x"]
            elif (
                style == "right"
                or (style == "right-left" and even)
                or (style == "left-right" and not even)
            ):
                df["xminv"] = df["x"]

            # Make sure it's sorted properly to draw the outline
            # i.e violin = kde + mirror kde,
            # bottom to top then top to bottom
            n = len(df)
            polygon_df = pd.concat(
                [df.sort_values("y"), df.sort_values("y", ascending=False)],
                axis=0,
                ignore_index=True,
            )

            _df = polygon_df.iloc
            _loc = polygon_df.columns.get_loc
            _df[:n, _loc("x")] = _df[:n, _loc("xminv")]
            _df[n:, _loc("x")] = _df[n:, _loc("xmaxv")]

            # Close the polygon: set first and last point the same
            polygon_df.loc[-1, :] = polygon_df.loc[0, :]

            # plot violin polygon
            geom_polygon.draw_group(
                polygon_df, panel_params, coord, ax, **params
            )

            if quantiles is not None:
                # Get dataframe with quantile segments and that
                # with aesthetics then put them together
                # Each quantile segment is defined by 2 points and
                # they all get similar aesthetics
                aes_df = df.drop(["x", "y", "group"], axis=1)
                aes_df.reset_index(inplace=True)
                idx = [0] * 2 * len(quantiles)
                aes_df = aes_df.iloc[idx, :].reset_index(drop=True)
                segment_df = pd.concat(
                    [make_quantile_df(df, quantiles), aes_df], axis=1
                )

                # plot quantile segments
                geom_path.draw_group(
                    segment_df, panel_params, coord, ax, **params
                )


def make_quantile_df(
    data: pd.DataFrame, draw_quantiles: FloatArray
) -> pd.DataFrame:
    """
    Return a dataframe with info needed to draw quantile segments
    """
    from scipy.interpolate import interp1d

    dens = data["density"].cumsum() / data["density"].sum()
    ecdf = interp1d(dens, data["y"], assume_sorted=True)
    ys = ecdf(draw_quantiles)

    # Get the violin bounds for the requested quantiles
    violin_xminvs = interp1d(data["y"], data["xminv"])(ys)
    violin_xmaxvs = interp1d(data["y"], data["xmaxv"])(ys)

    data = pd.DataFrame(
        {
            "x": interleave(violin_xminvs, violin_xmaxvs),
            "y": np.repeat(ys, 2),
            "group": np.repeat(np.arange(1, len(ys) + 1), 2),
        }
    )

    return data



from __future__ import annotations

from warnings import warn

import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineError, PlotnineWarning
from ..mapping.evaluation import after_stat
from ..utils import groupby_apply
from .binning import (
    assign_bins,
    breaks_from_bins,
    breaks_from_binwidth,
    freedman_diaconis_bins,
)
from .stat import stat


@document
class stat_bindot(stat):
    """
    Binning for a dot plot

    {usage}

    Parameters
    ----------
    {common_parameters}
    bins : int, optional (default: None)
        Number of bins. Overridden by binwidth. If :py:`None`,
        a number is computed using the freedman-diaconis method.
    binwidth : float, optional (default: None)
        When :py:`method='dotdensity'`, this specifies the maximum
        binwidth. When :py:`method='histodot'`, this specifies the
        binwidth. This supercedes the ``bins``.
    origin : float, optional (default: None)
        When :py:`method='histodot'`, origin of the first bin.
    width : float, optional (default: 0.9)
        When :py:`binaxis='y'`, the spacing of the dotstacks for
        dodging.
    binaxis : str, optional (default: x)
        Axis to bin along. Either :py:`'x'` or :py:`'y'`
    method : str, optional (default: dotdensity)
        One of *dotdensity* or *histodot*. These provide either of
        dot-density binning or fixed bin widths.
    binpositions : str, optional (default: bygroup)
        Position of the bins when :py:`method='dotdensity'`. The value
        is one of::

            'bygroup'  # positions of the bins for each group are
                       # determined separately.
            'all'      # positions of the bins are determined with all
                       # data taken together. This aligns the dots
                       # stacks across multiple groups.

    drop : bool, optional (default: False)
        If :py:`True`, remove all bins with zero counts.
    right : bool, optional (default: True)
        When :py:`method='histodot'`, :py:`True` means include right
        edge of the bins and if :py:`False` the left edge is included.
    breaks : array-like, optional (default: None)
        Bin boundaries for :py:`method='histodot'`. This supercedes the
        ``binwidth`` and ``bins``.

    See Also
    --------
    plotnine.stats.stat_bin
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

         'count'    # number of points in bin
         'density'  # density of points in bin, scaled to integrate to 1
         'ncount'   # count, scaled to maximum of 1
         'ndensity' # density, scaled to maximum of 1

    """

    REQUIRED_AES = {"x"}
    NON_MISSING_AES = {"weight"}
    DEFAULT_PARAMS = {
        "geom": "dotplot",
        "position": "identity",
        "na_rm": False,
        "bins": None,
        "binwidth": None,
        "origin": None,
        "width": 0.9,
        "binaxis": "x",
        "method": "dotdensity",
        "binpositions": "bygroup",
        "drop": False,
        "right": True,
        "breaks": None,
    }
    DEFAULT_AES = {"y": after_stat("count")}
    CREATES = {"width", "count", "density", "ncount", "ndensity"}

    def setup_params(self, data):
        params = self.params

        if (
            params["breaks"] is None
            and params["binwidth"] is None
            and params["bins"] is None
        ):
            params = params.copy()
            params["bins"] = freedman_diaconis_bins(data["x"])
            msg = (
                "'stat_bin()' using 'bins = {}'. "
                "Pick better value with 'binwidth'."
            )
            warn(msg.format(params["bins"]), PlotnineWarning)

        return params

    @classmethod
    def compute_panel(cls, data, scales, **params):
        if (
            params["method"] == "dotdensity"
            and params["binpositions"] == "all"
        ):
            binaxis = params["binaxis"]
            if binaxis == "x":
                newdata = densitybin(
                    x=data["x"],
                    weight=data.get("weight"),
                    binwidth=params["binwidth"],
                    bins=params["bins"],
                )
                data = data.sort_values("x")
                data.reset_index(inplace=True, drop=True)
                newdata = newdata.sort_values("x")
                newdata.reset_index(inplace=True, drop=True)
            elif binaxis == "y":
                newdata = densitybin(
                    x=data["y"],
                    weight=data.get("weight"),
                    binwidth=params["binwidth"],
                    bins=params["bins"],
                )
                data = data.sort_values("y")
                data.reset_index(inplace=True, drop=True)
                newdata = newdata.sort_values("x")
                newdata.reset_index(inplace=True, drop=True)
            else:
                raise ValueError(f"Unknown value {binaxis=}")

            data["bin"] = newdata["bin"]
            data["binwidth"] = newdata["binwidth"]
            data["weight"] = newdata["weight"]
            data["bincenter"] = newdata["bincenter"]
        return super(cls, stat_bindot).compute_panel(data, scales, **params)

    @classmethod
    def compute_group(cls, data, scales, **params):
        # Check that weights are whole numbers
        # (for dots, weights must be whole)
        weight: pd.Series | None = data.get("weight")  # pyright: ignore
        if weight is not None:
            int_status = [(w * 1.0).is_integer() for w in weight]
            if not all(int_status):
                raise PlotnineError(
                    "Weights for stat_bindot must be nonnegative integers."
                )

        if params["binaxis"] == "x":
            rangee = scales.x.dimension((0, 0))
            values = data["x"].values
            midline = 0  #  Make pyright happy
        else:
            rangee = scales.y.dimension((0, 0))
            values = data["y"].values
            # The middle of each group, on the stack axis
            midline = np.mean([data["x"].min(), data["x"].max()])

        if params["method"] == "histodot":
            if params["binwidth"] is not None:
                breaks = breaks_from_binwidth(
                    rangee, params["binwidth"], boundary=params["origin"]
                )
            else:
                breaks = breaks_from_bins(
                    rangee, params["bins"], boundary=params["origin"]
                )

            closed = "right" if params["right"] else "left"
            data = assign_bins(
                values, breaks, data.get("weight"), pad=False, closed=closed
            )
            # for consistency
            data.rename(
                columns={"width": "binwidth", "x": "bincenter"}, inplace=True
            )
        elif params["method"] == "dotdensity":
            # If bin centers are found by group instead of by all,
            # find the bin centers (If binpositions=="all", then
            # we'll already have bin centers.)
            if params["binpositions"] == "bygroup":
                data = densitybin(
                    x=values,
                    weight=weight,
                    binwidth=params["binwidth"],
                    bins=params["bins"],
                    rangee=rangee,
                )

            # Collapse each bin and get a count
            def func(df):
                return pd.DataFrame(
                    {
                        "binwidth": [df["binwidth"].iloc[0]],
                        "bincenter": [df["bincenter"].iloc[0]],
                        "count": [int(df["weight"].sum())],
                    }
                )

            # plyr::ddply + plyr::summarize
            data = groupby_apply(data, "bincenter", func)

            if data["count"].sum() != 0:
                data.loc[np.isnan(data["count"]), "count"] = 0
                data["ncount"] = data["count"] / data["count"].abs().max()
                if params["drop"]:
                    data = data[data["count"] > 0]
                    data.reset_index(inplace=True, drop=True)

        if params["binaxis"] == "x":
            data["x"] = data.pop("bincenter")
            # For x binning, the width of the geoms
            # is same as the width of the bin
            data["width"] = data["binwidth"]
        else:
            data["y"] = data.pop("bincenter")
            # For y binning, set the x midline.
            # This is needed for continuous x axis
            data["x"] = midline

        return data


def densitybin(x, weight=None, binwidth=None, bins=None, rangee=None):
    """
    Do density binning

    It does not collapse each bin with a count.

    Parameters
    ----------
    x : array-like
        Numbers to bin
    weight : array-like
        Weights
    binwidth : numeric
        Size of the bins
    bins : int
        Number of bins
    rangee : tuple
        Range of x

    Returns
    -------
    data : DataFrame
    """
    if all(pd.isnull(x)):
        return pd.DataFrame()

    if weight is None:
        weight = np.ones(len(x))
    weight = np.asarray(weight)
    weight[np.isnan(weight)] = 0

    if rangee is None:
        rangee = np.min(x), np.max(x)
    if bins is None:
        bins = 30
    if binwidth is None:
        binwidth = np.ptp(rangee) / bins

    # Sort weight and x, by x
    order = np.argsort(x)
    weight = weight[order]
    x = x[order]

    cbin = 0  # Current bin ID
    bin_ids = []  # The bin ID for each observation
    # End position of current bin (scan left to right)
    binend = -np.inf

    # Scan list and put dots in bins
    for value in x:
        # If past end of bin, start a new bin at this point
        if value >= binend:
            binend = value + binwidth
            cbin = cbin + 1
        bin_ids.append(cbin)

    def func(series):
        return (series.min() + series.max()) / 2

    results = pd.DataFrame(
        {
            "x": x,
            "bin": bin_ids,
            "binwidth": binwidth,
            "weight": weight,
        }
    )
    # This is a plyr::ddply
    results["bincenter"] = results.groupby("bin")["x"].transform(func)
    return results



import os
import subprocess
from pathlib import Path

import pytest

# Helps contributors catch linter errors
# when they run make test.

is_CI = os.environ.get("CI") is not None


@pytest.mark.skipif(is_CI, reason="Helps contributors catch linter errors")
def test_ruff():
    plotnine_dir = str(Path(__file__).parent.parent.absolute())
    p = subprocess.Popen(
        ["ruff", plotnine_dir],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    # Ignore the stderr msg as it is mostly noise
    # about coverage warnings
    stdout, _ = p.communicate()
    s = stdout.decode("utf-8")
    msg = f"ruff found the following issues: \n\n{s}"
    assert p.returncode == 0, msg


@pytest.mark.skipif(is_CI, reason="Helps contributors catch linter errors")
def test_black():
    plotnine_dir = str(Path(__file__).parent.parent.absolute())
    p = subprocess.Popen(
        ["black", plotnine_dir, "--check"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    # Ignore the stderr msg as it is mostly noise
    # about coverage warnings
    _, stderr = p.communicate()
    s = stderr.decode("utf-8")
    msg = f"black found the following issues: \n\n{s}"
    assert p.returncode == 0, msg



from __future__ import annotations

import typing
from copy import deepcopy

import numpy as np

from ..utils import jitter, resolution
from .position import position

if typing.TYPE_CHECKING:
    from plotnine.typing import FloatArray, FloatArrayLike


class position_jitter(position):
    """
    Jitter points to avoid overplotting

    Parameters
    ----------
    width : float
        Proportion to jitter in horizontal direction.
        Default is ``0.4`` of the resolution of the data.
    height : float
        Proportion to jitter in vertical direction.
        Default is ``0.4`` of the resolution of the data.
    random_state : int or ~numpy.random.RandomState, optional
        Seed or Random number generator to use. If ``None``, then
        numpy global generator :class:`numpy.random` is used.
    """

    REQUIRED_AES = {"x", "y"}

    def __init__(self, width=None, height=None, random_state=None):
        self.params = {
            "width": width,
            "height": height,
            "random_state": random_state,
        }

    def setup_params(self, data):
        params = deepcopy(self.params)
        if params["width"] is None:
            params["width"] = resolution(data["x"]) * 0.4
        if params["height"] is None:
            params["height"] = resolution(data["y"]) * 0.4
        if not params["random_state"]:
            params["random_state"] = np.random
        return params

    @classmethod
    def compute_layer(cls, data, params, layout):
        trans_x = None  # pyright: ignore
        trans_y = None  # pyright: ignore

        if params["width"]:

            def trans_x(x: FloatArrayLike) -> FloatArray:
                return jitter(
                    x,
                    amount=params["width"],
                    random_state=params["random_state"],
                )

        if params["height"]:

            def trans_y(y):
                return jitter(
                    y,
                    amount=params["height"],
                    random_state=params["random_state"],
                )

        return cls.transform_position(data, trans_x, trans_y)



import numpy as np
import pandas as pd
import pytest

from plotnine import aes, arrow, ggplot, stat_function
from plotnine.exceptions import PlotnineError

n = 10
df = pd.DataFrame({"x": range(1, n + 1)})


def test_limits():
    p = (
        ggplot(df, aes("x"))
        + stat_function(
            fun=np.cos, size=2, color="blue", arrow=arrow(ends="first")
        )
        + stat_function(
            fun=np.cos,
            xlim=(10, 20),
            size=2,
            color="red",
            arrow=arrow(ends="last"),
        )
    )
    assert p == "limits"


def test_args():
    def fun(x, f=lambda x: x, mul=1, add=0):
        return f(x) * mul + add

    # no args, single arg, tuple of args, dict of args
    p = (
        ggplot(df, aes("x"))
        + stat_function(fun=fun, size=2, color="blue")
        + stat_function(fun=fun, size=2, color="red", args=np.cos)
        + stat_function(fun=fun, size=2, color="green", args=(np.cos, 2, 1))
        + stat_function(
            fun=fun,
            size=2,
            color="purple",
            args={"f": np.cos, "mul": 3, "add": 2},
        )
    )

    assert p == "args"


def test_exceptions():
    # no mapping, no limits
    p = ggplot(df) + stat_function(fun=np.sin)
    p.draw_test()

    # fun not callable
    with pytest.raises(PlotnineError):
        p = ggplot(df, aes("x"))
        print(p + stat_function(fun=1))



import os

import numpy as np
import pandas as pd
import pytest

from plotnine import (
    aes,
    geom_label,
    geom_point,
    geom_text,
    ggplot,
    scale_size_continuous,
    scale_y_continuous,
)
from plotnine.data import mtcars
from plotnine.exceptions import PlotnineWarning

is_CI = os.environ.get("CI") is not None

n = 5
labels = [
    "ggplot",
    "aesthetics",
    "data",
    "geoms",
    r"$\mathbf{statistics^2}$",
    "scales",
    "coordinates",
]
df = pd.DataFrame(
    {
        "x": [1] * n,
        "y": range(n),
        "label": labels[:n],
        "z": range(n),
        "angle": np.linspace(0, 90, n),
    }
)

adjust_text = {
    "expand_points": (2, 2),
    "arrowprops": {"arrowstyle": "->", "color": "red"},
}


def test_text_aesthetics():
    p = (
        ggplot(df, aes(y="y", label="label"))
        + geom_text(aes("x", label="label"), size=15, ha="left")
        + geom_text(
            aes("x+1", angle="angle"), size=15, va="top", show_legend=False
        )
        + geom_text(
            aes("x+2", label="label", alpha="z"), size=15, show_legend=False
        )
        + geom_text(aes("x+3", color="factor(z)"), size=15, show_legend=False)
        + geom_text(aes("x+5", size="z"), ha="right", show_legend=False)
        + scale_size_continuous(range=(12, 30))
        + scale_y_continuous(limits=(-0.5, n - 0.5))
    )

    assert p == "text_aesthetics"


def test_label_aesthetics():
    p = (
        ggplot(df, aes(y="y", label="label"))
        + geom_label(
            aes("x", label="label", fill="z"),
            size=15,
            ha="left",
            show_legend=False,
            boxcolor="red",
        )
        + geom_label(
            aes("x+1", angle="angle"), size=15, va="top", show_legend=False
        )
        + geom_label(
            aes("x+2", label="label", alpha="z"), size=15, show_legend=False
        )
        + geom_label(aes("x+3", color="factor(z)"), size=15, show_legend=False)
        + geom_label(aes("x+5", size="z"), ha="right", show_legend=False)
        + scale_size_continuous(range=(12, 30))
        + scale_y_continuous(limits=(-0.5, n - 0.5))
    )

    assert p == "label_aesthetics"


@pytest.mark.skipif(is_CI, reason="Sudden small difference on GHA")
def test_adjust_text():
    p = (
        ggplot(mtcars.tail(2), aes("mpg", "disp", label="name"))
        + geom_point(size=5, fill="black")
        + geom_text(adjust_text=adjust_text)
    )
    assert p == "adjust_text"


@pytest.mark.skipif(is_CI, reason="Sudden small difference on GHA")
def test_adjust_label():
    p = (
        ggplot(mtcars.tail(2), aes("mpg", "disp", label="name"))
        + geom_point(size=5, fill="black")
        + geom_label(adjust_text=adjust_text)
    )
    assert p == "adjust_label"


@pytest.mark.skipif(is_CI, reason="Sudden small difference on GHA")
def test_adjust_text_default_color():
    adjust_text2 = adjust_text.copy()
    del adjust_text2["arrowprops"]["color"]

    p = (
        ggplot(mtcars.tail(2), aes("mpg", "disp", label="name"))
        + aes(color="factor(cyl)")
        + geom_point(size=5, fill="black")
        + geom_text(adjust_text=adjust_text2)
    )
    assert p == "adjust_text_default_color"


def test_format_missing_values():
    df = pd.DataFrame(
        {
            "x": [1, 2, 3, 4],
            "y": [1, 2, 3, 4],
            "c1": [1.1, 2.2, None, 4],
            "c2": ["1.1", "2.2", None, (4, 0)],
        }
    )
    p = (
        ggplot(df, aes("x", "y"))
        + geom_point()
        + geom_text(
            aes(label="c1"),
            nudge_y=0.03,
            va="bottom",
            color="blue",
            format_string="{}",
        )
        + geom_text(
            aes(label="c2"),
            nudge_y=-0.03,
            va="top",
            color="red",
            format_string="{!r}",
        )
    )
    with pytest.warns(PlotnineWarning):
        assert p == "format_missing_values"



import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineError
from ..utils import get_valid_kwargs, uniquecols
from .stat import stat


def bootstrap_statistics(
    series,
    statistic,
    n_samples=1000,
    confidence_interval=0.95,
    random_state=None,
):
    """
    Default parameters taken from
    R's Hmisc smean.cl.boot
    """
    if random_state is None:
        random_state = np.random

    alpha = 1 - confidence_interval
    size = (n_samples, len(series))
    inds = random_state.randint(0, len(series), size=size)
    samples = series.values[inds]
    means = np.sort(statistic(samples, axis=1))
    return pd.DataFrame(
        {
            "ymin": means[int((alpha / 2) * n_samples)],
            "ymax": means[int((1 - alpha / 2) * n_samples)],
            "y": [statistic(series)],
        }
    )


def mean_cl_boot(
    series, n_samples=1000, confidence_interval=0.95, random_state=None
):
    """
    Bootstrapped mean with confidence interval

    Parameters
    ----------
    series : pandas.Series
        Values
    n_samples : int (default: 1000)
        Number of sample to draw.
    confidence_interval : float
        Confidence interval in the range (0, 1).
    random_state : int or ~numpy.random.RandomState, optional
        Seed or Random number generator to use. If ``None``, then
        numpy global generator :class:`numpy.random` is used.
    """
    return bootstrap_statistics(
        series,
        np.mean,
        n_samples=n_samples,
        confidence_interval=confidence_interval,
        random_state=random_state,
    )


def mean_cl_normal(series, confidence_interval=0.95):
    """
    Mean with confidence interval assuming normal distribution

    Credit: from http://stackoverflow.com/a/15034143

    Parameters
    ----------
    series : pandas.Series
        Values
    confidence_interval : float
        Confidence interval in the range (0, 1).
    """
    import scipy.stats as stats

    a = np.asarray(series)
    m = np.mean(a)
    se = stats.sem(a)
    h = se * stats.t._ppf((1 + confidence_interval) / 2, len(a) - 1)
    return pd.DataFrame({"y": [m], "ymin": m - h, "ymax": m + h})


def mean_sdl(series, mult=2):
    """
    Mean +/- a constant times the standard deviation

    Parameters
    ----------
    series : pandas.Series
        Values
    mult : float
        Multiplication factor.
    """
    m = series.mean()
    s = series.std()
    return pd.DataFrame({"y": [m], "ymin": m - mult * s, "ymax": m + mult * s})


def median_hilow(series, confidence_interval=0.95):
    """
    Median and a selected pair of outer quantiles having equal tail areas

    Parameters
    ----------
    series : pandas.Series
        Values
    confidence_interval : float
        Confidence interval in the range (0, 1).
    """
    tail = (1 - confidence_interval) / 2
    return pd.DataFrame(
        {
            "y": [np.median(series)],
            "ymin": np.percentile(series, 100 * tail),
            "ymax": np.percentile(series, 100 * (1 - tail)),
        }
    )


def mean_se(series, mult=1):
    """
    Calculate mean and standard errors on either side

    Parameters
    ----------
    series : pandas.Series
        Values
    mult : float
        Multiplication factor.
    """
    m = np.mean(series)
    se = mult * np.sqrt(np.var(series) / len(series))
    return pd.DataFrame({"y": [m], "ymin": m - se, "ymax": m + se})


function_dict = {
    "mean_cl_boot": mean_cl_boot,
    "mean_cl_normal": mean_cl_normal,
    "mean_sdl": mean_sdl,
    "median_hilow": median_hilow,
    "mean_se": mean_se,
}


def make_summary_fun(fun_data, fun_y, fun_ymin, fun_ymax, fun_args):
    """
    Make summary function
    """
    if isinstance(fun_data, str):
        fun_data = function_dict[fun_data]

    if any([fun_y, fun_ymin, fun_ymax]):

        def func(df) -> pd.DataFrame:
            d = {}
            if fun_y:
                kwargs = get_valid_kwargs(fun_y, fun_args)
                d["y"] = [fun_y(df["y"], **kwargs)]
            if fun_ymin:
                kwargs = get_valid_kwargs(fun_ymin, fun_args)
                d["ymin"] = [fun_ymin(df["y"], **kwargs)]
            if fun_ymax:
                kwargs = get_valid_kwargs(fun_ymax, fun_args)
                d["ymax"] = [fun_ymax(df["y"], **kwargs)]
            return pd.DataFrame(d)

    elif fun_data:
        kwargs = get_valid_kwargs(fun_data, fun_args)

        def func(df) -> pd.DataFrame:
            return fun_data(df["y"], **kwargs)

    else:
        raise ValueError(f"Bad value for function fun_data={fun_data}")

    return func


@document
class stat_summary(stat):
    """
    Calculate summary statistics depending on x

    {usage}

    Parameters
    ----------
    {common_parameters}
    fun_data : str or function, optional
        One of::

            # Bootstrapped mean, confidence interval
            # Arguments:
            #     n_samples - No. of samples to draw
            #     confidence_interval
            #     random_state
            'mean_cl_boot'

            # Mean, C.I. assuming normal distribution
            # Arguments:
            #     confidence_interval
            'mean_cl_normal'

            # Mean, standard deviation * constant
            # Arguments:
            #     mult - multiplication factor
            'mean_sdl'

            # Median, outlier quantiles with equal tail areas
            # Arguments:
            #     confidence_interval
            'median_hilow'

            # Mean, Standard Errors * constant
            # Arguments:
            #     mult - multiplication factor
            'mean_se'

        or any function that takes a array and returns a dataframe
        with three columns named ``y``, ``ymin`` and ``ymax``.
        Defaults to :py:`'mean_cl_boot'`.
    fun_y : function, optional (default: None)
        Any function that takes a array-like and returns a value
        fun_ymin : function (default:None)
        Any function that takes an array-like and returns a value
    fun_ymax : function, optional (default: None)
        Any function that takes an array-like and returns a value
    fun_args : dict, optional (default: None)
        Arguments to any of the functions. Provided the names of the
        arguments of the different functions are in not conflict, the
        arguments will be assigned to the right functions. If there is
        a conflict, create a wrapper function that resolves the
        ambiguity in the argument names.
    random_state : int or ~numpy.random.RandomState, optional
        Seed or Random number generator to use. If ``None``, then
        numpy global generator :class:`numpy.random` is used.

    Notes
    -----
    If any of ``fun_y``, ``fun_ymin`` or ``fun_ymax`` are provided, the
    value of ``fun_data`` will be ignored.

    See Also
    --------
    plotnine.geoms.geom_pointrange
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

        'ymin'  # ymin computed by the summary function
        'ymax'  # ymax computed by the summary function

    Calculated aesthetics are accessed using the `after_stat` function.
    e.g. :py:`after_stat('ymin')`.
    """

    REQUIRED_AES = {"x", "y"}
    DEFAULT_PARAMS = {
        "geom": "pointrange",
        "position": "identity",
        "na_rm": False,
        "fun_data": "mean_cl_boot",
        "fun_y": None,
        "fun_ymin": None,
        "fun_ymax": None,
        "fun_args": None,
        "random_state": None,
    }
    CREATES = {"ymin", "ymax"}

    def setup_params(self, data):
        keys = ("fun_data", "fun_y", "fun_ymin", "fun_ymax")
        if not any(self.params[k] for k in keys):
            raise PlotnineError("No summary function")

        if self.params["fun_args"] is None:
            self.params["fun_args"] = {}

        if "random_state" not in self.params["fun_args"]:
            if self.params["random_state"]:
                random_state = self.params["random_state"]
                if random_state is None:
                    random_state = np.random
                elif isinstance(random_state, int):
                    random_state = np.random.RandomState(random_state)

                self.params["fun_args"]["random_state"] = random_state

        return self.params

    @classmethod
    def compute_panel(cls, data, scales, **params):
        func = make_summary_fun(
            params["fun_data"],
            params["fun_y"],
            params["fun_ymin"],
            params["fun_ymax"],
            params["fun_args"],
        )

        # break a dataframe into pieces, summarise each piece,
        # and join the pieces back together, retaining original
        # columns unaffected by the summary.
        summaries = []
        for (group, x), df in data.groupby(["group", "x"]):
            summary = func(df)
            summary["x"] = x
            summary["group"] = group
            unique = uniquecols(df)
            if "y" in unique:
                unique = unique.drop("y", axis=1)
            merged = summary.merge(unique, on=["group", "x"])
            summaries.append(merged)

        new_data = pd.concat(summaries, axis=0, ignore_index=True)
        return new_data



import numpy as np
import pandas as pd

from ..doctools import document
from ..mapping.evaluation import after_stat
from .density import get_var_type, kde
from .stat import stat


@document
class stat_pointdensity(stat):
    """
    Compute density estimation for each point

    {usage}

    Parameters
    ----------
    {common_parameters}
    package : str in ``['statsmodels', 'scipy', 'sklearn']``
        Package whose kernel density estimation to use. Default is
        statsmodels.
    kde_params : dict
        Keyword arguments to pass on to the kde class.

    See Also
    --------
    statsmodels.nonparametric.kde.KDEMultivariate
    scipy.stats.gaussian_kde
    sklearn.neighbors.KernelDensity
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

        'density'   # Computed density at a point

    """
    REQUIRED_AES = {"x", "y"}
    DEFAULT_AES = {"color": after_stat("density")}
    DEFAULT_PARAMS = {
        "geom": "density_2d",
        "position": "identity",
        "na_rm": False,
        "package": "statsmodels",
        "kde_params": None,
    }
    CREATES = {"density"}

    def setup_params(self, data):
        params = self.params.copy()
        if params["kde_params"] is None:
            params["kde_params"] = {}

        kde_params = params["kde_params"]
        if params["package"] == "statsmodels":
            params["package"] = "statsmodels-m"
            if "var_type" not in kde_params:
                kde_params["var_type"] = "{}{}".format(
                    get_var_type(data["x"]), get_var_type(data["y"])
                )

        return params

    @classmethod
    def compute_group(cls, data, scales, **params):
        package = params["package"]
        kde_params = params["kde_params"]

        var_data = np.array([data["x"].values, data["y"].values]).T
        density = kde(var_data, var_data, package, **kde_params)

        data = pd.DataFrame(
            {
                "x": data["x"],
                "y": data["y"],
                "density": density.flatten(),
            }
        )

        return data



import scipy.stats as stats

from ..exceptions import PlotnineError


def _hasattrs(obj, attrs):
    return all(hasattr(obj, attr) for attr in attrs)


# Continuous univariate
continuous = {
    k for k in dir(stats) if _hasattrs(getattr(stats, k), ("pdf", "cdf"))
}

# Discrete univariate
discrete = {k for k in dir(stats) if hasattr(getattr(stats, k), "pmf")}

univariate = continuous | discrete


def get(name):
    """
    Get any scipy.stats distribution of a given name
    """
    try:
        return getattr(stats, name)
    except AttributeError:
        raise PlotnineError(f"Unknown distribution '{name}'")


def get_continuous_distribution(name):
    """
    Get continuous scipy.stats distribution of a given name
    """
    if name not in continuous:
        msg = "Unknown continuous distribution '{}'"
        raise ValueError(msg.format(name))

    return getattr(stats, name)


def get_univariate(name):
    """
    Get univariate scipy.stats distribution of a given name
    """
    if name not in univariate:
        msg = "Unknown univariate distribution '{}'"
        raise ValueError(msg.format(name))

    return get(name)



from __future__ import annotations

import typing

from ..doctools import document
from .geom import geom
from .geom_linerange import geom_linerange
from .geom_path import geom_path
from .geom_point import geom_point

if typing.TYPE_CHECKING:
    from typing import Any

    import pandas as pd

    from plotnine.iapi import panel_view
    from plotnine.typing import Axes, Coord, DrawingArea, Layer


@document
class geom_pointrange(geom):
    """
    Vertical interval represented by a line with a point

    {usage}

    Parameters
    ----------
    {common_parameters}
    fatten : float, optional (default: 2)
        A multiplicative factor used to increase the size of the
        point along the line-range.
    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": "black",
        "fill": None,
        "linetype": "solid",
        "shape": "o",
        "size": 0.5,
    }
    REQUIRED_AES = {"x", "y", "ymin", "ymax"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "fatten": 4,
    }

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        geom_linerange.draw_group(
            data.copy(), panel_params, coord, ax, **params
        )
        data["size"] = data["size"] * params["fatten"]
        data["stroke"] = geom_point.DEFAULT_AES["stroke"]
        geom_point.draw_group(data, panel_params, coord, ax, **params)

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: Layer
    ) -> DrawingArea:
        """
        Draw a point in the box

        Parameters
        ----------
        data : Series
            Data Row
        da : DrawingArea
            Canvas
        lyr : layer
            Layer

        Returns
        -------
        out : DrawingArea
        """
        geom_path.draw_legend(data, da, lyr)
        data["size"] = data["size"] * lyr.geom.params["fatten"]
        data["stroke"] = geom_point.DEFAULT_AES["stroke"]
        geom_point.draw_legend(data, da, lyr)
        return da



from __future__ import annotations

import typing

import numpy as np
from mizani.bounds import expand_range_distinct

from ..iapi import range_view
from ..utils import ignore_warnings

if typing.TYPE_CHECKING:
    from typing import Type

    from plotnine.typing import (
        CoordRange,
        Trans,
        TupleFloat2,
        TupleFloat4,
    )


def _expand_range_distinct(
    x: TupleFloat2,
    expand: TupleFloat2 | TupleFloat4,
) -> TupleFloat2:
    # Expand ascending and descending order range
    a, b = x
    if a > b:
        b, a = expand_range_distinct((b, a), expand)
    else:
        a, b = expand_range_distinct((a, b), expand)
    return (a, b)


def expand_range(
    x: CoordRange,
    expand: TupleFloat2 | TupleFloat4,
    trans: Trans | Type[Trans],
) -> range_view:
    """
    Expand Coordinate Range in coordinate space

    Parameters
    ----------
    x:
        (max, min) in data scale
    expand:
        How to expand
    trans:
        Coordinate transformation
    """
    x_coord_space = trans.transform(x)
    x_coord = _expand_range_distinct(x_coord_space, expand)

    with ignore_warnings(RuntimeWarning):
        # Consequences of the runtimewarning (NaNs and infs)
        # are dealt with below
        final_x = trans.inverse(x_coord)

    l0, l1 = x
    f0, f1 = final_x
    final_range = (
        f0 if np.isfinite(f0) else l0,
        f1 if np.isfinite(f1) else l1,
    )

    ranges = range_view(range=final_range, range_coord=x_coord)
    return ranges



from ..doctools import document
from .geom_rect import geom_rect


@document
class geom_bin_2d(geom_rect):
    """
    Heatmap of 2d bin counts

    Divides the plane into rectangles, counts the number of
    cases in each rectangle, and then (by default) maps the number
    of cases to the rectangle's fill. This is a useful alternative
    to geom_point in the presence of overplotting.

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_PARAMS = {"stat": "bin_2d", "position": "identity", "na_rm": False}


geom_bin2d = geom_bin_2d



from warnings import warn

import numpy as np

from ..doctools import document
from ..exceptions import PlotnineWarning
from ..utils import alias
from .scale import scale_continuous, scale_discrete


@document
class scale_stroke_continuous(scale_continuous):
    """
    Continuous Stroke Scale

    Parameters
    ----------
    range : array_like
        Range ([Minimum, Maximum]) of output stroke values.
        Should be between 0 and 1. Default is ``(1, 6)``
    {superclass_parameters}
    """

    _aesthetics = ["stroke"]

    def __init__(self, range=(1, 6), **kwargs):
        from mizani.palettes import rescale_pal

        # TODO: fix types in mizani
        self.palette = rescale_pal(range)  # pyright: ignore
        scale_continuous.__init__(self, **kwargs)


@document
class scale_stroke_ordinal(scale_discrete):
    """
    Discrete Stroke Scale

    Parameters
    ----------
    range : array_like
        Range ([Minimum, Maximum]) of output stroke values.
        Should be between 0 and 1. Default is ``(1, 6)``
    {superclass_parameters}
    """

    _aesthetics = ["stroke"]

    def __init__(self, range=(1, 6), **kwargs):
        def palette(n):
            return np.linspace(range[0], range[1], n)

        self.palette = palette
        scale_discrete.__init__(self, **kwargs)


@document
class scale_stroke_discrete(scale_stroke_ordinal):
    """
    Discrete Stroke Scale

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["stroke"]

    def __init__(self, **kwargs):
        warn(
            "Using stroke for a ordinal variable is not advised.",
            PlotnineWarning,
        )
        super().__init__(self, **kwargs)


alias("scale_stroke", scale_stroke_continuous)



from warnings import warn

from ..doctools import document
from ..exceptions import PlotnineError, PlotnineWarning
from ..utils import alias
from .scale import scale_continuous, scale_discrete

LINETYPES = ["solid", "dashed", "dashdot", "dotted"]


@document
class scale_linetype(scale_discrete):
    """
    Scale for line patterns

    Parameters
    ----------
    {superclass_parameters}

    Notes
    -----
    The available linetypes are
    ``'solid', 'dashed', 'dashdot', 'dotted'``
    If you need more custom linetypes, use
    :class:`~plotnine.scales.scale_linetype_manual`
    """

    _aesthetics = ["linetype"]

    def __init__(self, **kwargs):
        from mizani.palettes import manual_pal

        self.palette = manual_pal(LINETYPES)
        super().__init__(**kwargs)


@document
class scale_linetype_ordinal(scale_linetype):
    """
    Scale for line patterns

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["linetype"]

    def __init__(self, **kwargs):
        warn(
            "Using linetype for an ordinal variable is not advised.",
            PlotnineWarning,
        )
        super().__init__(**kwargs)


class scale_linetype_continuous(scale_continuous):
    """
    Linetype scale
    """

    def __init__(self):
        raise PlotnineError(
            "A continuous variable can not be mapped to linetype"
        )


alias("scale_linetype_discrete", scale_linetype)



from __future__ import annotations

import typing
from copy import copy

import numpy as np
import pandas as pd

from ..exceptions import PlotnineError
from ..utils import groupby_apply, pivot_apply
from .position_dodge import position_dodge

if typing.TYPE_CHECKING:
    from plotnine.typing import IntArray


class position_dodge2(position_dodge):
    """
    Dodge overlaps and place objects side-by-side

    This is an enhanced version of
    :class:`~plotnine.positions.position_dodge` that can deal
    with rectangular overlaps that do not share a lower x border.

    Parameters
    ----------
    width: float
        Dodging width, when different to the width of the
        individual elements. This is useful when you want
        to align narrow geoms with wider geoms
    preserve: str in ``['total', 'single']``
        Should dodging preserve the total width of all elements
        at a position, or the width of a single element?
    padding : float
        Padding between elements at the same position.
        Elements are shrunk by this proportion to allow space
        between them (Default: 0.1)
    reverse : bool
        Reverse the default ordering of the groups. This is
        useful if you're rotating both the plot and legend.
        (Default: False)
    """

    REQUIRED_AES = {"x"}

    def __init__(
        self, width=None, preserve="total", padding=0.1, reverse=False
    ):
        self.params = {
            "width": width,
            "preserve": preserve,
            "padding": padding,
            "reverse": reverse,
        }

    def setup_params(self, data):
        if (
            ("xmin" not in data)
            and ("xmax" not in data)
            and (self.params["width"] is None)
        ):
            msg = "Width not defined. " "Set with `position_dodge2(width = ?)`"
            raise PlotnineError(msg)

        params = copy(self.params)

        if params["preserve"] == "total":
            params["n"] = None
        elif "x" in data:

            def max_x_values(gdf):
                n = gdf["x"].value_counts().max()
                return pd.DataFrame({"n": [n]})

            res = groupby_apply(data, "PANEL", max_x_values)
            params["n"] = res["n"].max()
        else:

            def _find_x_overlaps(gdf):
                return pd.DataFrame({"n": find_x_overlaps(gdf)})

            # interval geoms
            res = groupby_apply(data, "PANEL", _find_x_overlaps)
            params["n"] = res["n"].max()
        return params

    @classmethod
    def compute_panel(cls, data, scales, params):
        return cls.collide2(data, params=params)

    @staticmethod
    def strategy(data, params):
        padding = params["padding"]
        n = params["n"]

        if not all(col in data.columns for col in ["xmin", "xmax"]):
            data["xmin"] = data["x"]
            data["xmax"] = data["x"]

        # Groups of boxes that share the same position
        data["xid"] = find_x_overlaps(data)

        # Find newx using xid, i.e. the center of each group of
        # overlapping elements. for boxes, bars, etc. this should
        # be the same as original x, but for arbitrary rects it
        # may not be
        res1 = pivot_apply(data, "xmin", "xid", np.min)
        res2 = pivot_apply(data, "xmax", "xid", np.max)
        data["newx"] = (res1 + res2)[data["xid"].values].values / 2

        if n is None:
            # If n is None, preserve total widths of elements at
            # each position by dividing widths by the number of
            # elements at that position
            n = data["xid"].value_counts().values
            n = n[data.loc[:, "xid"] - 1]
            data["new_width"] = (data["xmax"] - data["xmin"]) / n
        else:
            data["new_width"] = (data["xmax"] - data["xmin"]) / n

        # Find the total width of each group of elements
        def sum_new_width(gdf):
            return pd.DataFrame(
                {
                    "size": [gdf["new_width"].sum()],
                    "newx": gdf["newx"].iloc[0],
                }
            )

        group_sizes = groupby_apply(data, "newx", sum_new_width)

        # Starting xmin for each group of elements
        starts = group_sizes["newx"] - (group_sizes["size"] / 2)

        # Set the elements in place
        for i, start in enumerate(starts, start=1):
            bool_idx = data["xid"] == i
            divisions = np.cumsum(
                np.hstack([start, data.loc[bool_idx, "new_width"]])
            )
            data.loc[bool_idx, "xmin"] = divisions[:-1]
            data.loc[bool_idx, "xmax"] = divisions[1:]

        # x values get moved to between xmin and xmax
        data["x"] = (data["xmin"] + data["xmax"]) / 2

        # Shrink elements to add space between them
        if data["xid"].duplicated().any():
            pad_width = data["new_width"] * (1 - padding)
            data["xmin"] = data["x"] - pad_width / 2
            data["xmax"] = data["x"] + pad_width / 2

        data = data.drop(columns=["xid", "newx", "new_width"], errors="ignore")
        return data


def find_x_overlaps(df: pd.DataFrame) -> IntArray:
    """
    Find overlapping regions along the x axis
    """
    n = len(df)
    overlaps = np.zeros(n, dtype=int)
    overlaps[0] = 1
    counter = 1
    for i in range(1, n):
        if df["xmin"].iloc[i] >= df["xmax"].iloc[i - 1]:
            counter += 1
        overlaps[i] = counter
    return overlaps



import pandas as pd

from plotnine import aes, geom_count, ggplot

df = pd.DataFrame(
    {
        "x": list("aaaaaaaaaabbbbbbbbbbcccccccccc"),
        "y": [
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            1,
            1,
            1,
            1,
            1,
            6,
            6,
            8,
            10,
            10,
            1,
            1,
            2,
            4,
            4,
            4,
            4,
            9,
            9,
            9,
        ],
    }
)


def test_discrete_x():
    p = ggplot(df, aes("x", "y")) + geom_count()

    assert p == "discrete_x"


def test_discrete_y():
    p = ggplot(df, aes("y", "x")) + geom_count()

    assert p == "discrete_y"


def test_continuous_x_y():
    p = ggplot(df, aes("y", "y")) + geom_count()

    assert p == "continuous_x_y"



from .elements import element_blank
from .theme import theme
from .theme_bw import theme_bw


class theme_minimal(theme_bw):
    """
    A minimalistic theme with no background annotations

    Parameters
    ----------
    base_size : int, optional
        Base font size. All text sizes are a scaled versions of
        the base font size. Default is 11.
    base_family : str, optional
        Base font family.
    """

    def __init__(self, base_size=11, base_family=None):
        super().__init__(base_size, base_family)
        self += theme(
            axis_ticks=element_blank(),
            legend_background=element_blank(),
            legend_key=element_blank(),
            panel_background=element_blank(),
            panel_border=element_blank(),
            plot_background=element_blank(),
            strip_background=element_blank(),
        )



import numpy as np
from geopandas import GeoDataFrame
from shapely.geometry import (
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)

from plotnine import aes, facet_wrap, geom_map, ggplot, labs, theme


def test_geometries():
    # Points
    points = [Point(0, 0), Point(0, 1), Point(1, 1), Point(1, 0)]
    point_names = [f"point{i}" for i in range(len(points))]
    df_point = GeoDataFrame({"names": point_names, "geometry": points})

    # LineString
    n = 5
    x = np.repeat(np.linspace(0, 1, n), 2)
    y = np.tile([0.375, 0.625], n)
    lines = [LineString(list(zip(x, y)))]
    df_line = GeoDataFrame({"name": ["line1"], "geometry": lines})

    # MultiLineString
    n = 5
    x = np.repeat(np.linspace(0, 1, n), 2)
    y = np.tile([0.375, 0.625], n) + 1
    line = list(zip(x, y))
    mlines = [MultiLineString([line[:5], line[5:]])]
    df_multiline = GeoDataFrame({"name": "multiline1", "geometry": mlines})

    #  Polygon
    polygons = [
        Polygon([(0.25, -0.25), (0.25, 0.25), (0.75, 0.25), (0.75, -0.25)]),
        Polygon([(0.25, 0.75), (0.75, 0.75), (0.5, 1.25)]),
    ]
    names = [f"polygon{i}" for i in range(len(polygons))]
    df_polygon = GeoDataFrame({"name": names, "geometry": polygons})

    p = (
        ggplot()
        + aes(fill="geometry.bounds.miny")
        + geom_map(df_polygon)
        + geom_map(df_point, size=4)
        + geom_map(df_line, size=2)
        + geom_map(df_multiline, size=2)
        + labs(fill="miny")
    )

    assert p == "geometries"


def test_multipolygon():
    # 2 MultiPolygon
    # 1. 4 Solid Squares
    # 2. 4 Squares with holes (to the upper-right of each of
    #    the squares in 1)
    length = 0.5
    centers = np.array([[1, 1], [1, 2], [2, 2], [2, 1]])
    corners = np.array([[-1, -1], [-1, 1], [1, 1], [1, -1]])
    shift = corners * (length / 2)
    shift_holes = corners * (length / 4)
    mpolygons = [
        MultiPolygon([(c + shift, None) for c in centers]),
        MultiPolygon(
            [(c + length + shift, [c + length + shift_holes]) for c in centers]
        ),
    ]
    names = [f"mpolygon{i}" for i in range(len(mpolygons))]
    df = GeoDataFrame({"name": names, "geometry": mpolygons})
    p = (
        ggplot()
        + aes(fill="geometry.bounds.miny")
        + geom_map(df)
        + labs(fill="miny")
        + theme(aspect_ratio=1)
    )

    assert p == "multipolygon"


def test_multipoint():
    mpoints = [
        MultiPoint([[0.0, 0.0], [1.0, 1.0]]),
        MultiPoint([[0.0, 1.0], [1.0, 2.0]]),
    ]

    mpoint_names = [f"mpoint{i}" for i in range(len(mpoints))]
    df = GeoDataFrame({"names": mpoint_names, "geometry": mpoints})

    p = (
        ggplot()
        + aes(fill="geometry.bounds.miny")
        + geom_map(df, size=5, color="None")
        + labs(fill="miny")
    )
    assert p == "multipoint"


def test_facet_wrap():
    #  Polygon
    polygons = [
        Polygon([(0.25, -0.25), (0.25, 0.25), (0.75, 0.25), (0.75, -0.25)]),
        Polygon([(0.25, 0.75), (0.75, 0.75), (0.5, 1.25)]),
    ]
    names = [f"polygon{i}" for i in range(len(polygons))]
    df_polygon = GeoDataFrame(
        {
            "name": names,
            "geometry": polygons,
            "shape": ["rectangle", "triangle"],
        }
    )

    p = (
        ggplot()
        + aes(fill="geometry.bounds.miny")
        + geom_map(df_polygon)
        + facet_wrap("shape")
        + labs(fill="miny")
    )
    assert p == "facet_wrap"



import numpy as np
import pytest

from plotnine import aes, geom_bar, ggplot
from plotnine.data import mtcars
from plotnine.exceptions import PlotnineError, PlotnineWarning
from plotnine.geoms.geom import geom
from plotnine.stats.stat import stat


def test_stat_basics():
    class stat_abc(stat):
        DEFAULT_PARAMS = {"geom": "point", "position": "identity"}
        CREATES = {"fill"}

    class stat_efg(stat):
        DEFAULT_PARAMS = {"geom": "point", "position": "identity"}
        REQUIRED_AES = {"weight"}
        CREATES = {"fill"}

    gg = ggplot(aes(x="wt", y="mpg"), mtcars)

    # stat_abc has no _calculate method
    with pytest.raises(NotImplementedError):
        print(gg + stat_abc())

    # stat_efg requires 'weight' aesthetic
    with pytest.raises(PlotnineError):
        print(gg + stat_efg())


def test_stat_parameter_sharing():
    # When the stat has a parameter with the same name as
    # the geom aesthetic,they both get their value

    # NOTE: This test may need to be modified when the
    # geom & stat internals change
    class stat_abc(stat):
        DEFAULT_PARAMS = {"geom": "point", "position": "identity", "weight": 1}
        REQUIRED_AES = {"x"}
        CREATES = {"y"}

        @classmethod
        def compute_panel(cls, data, scales, **params):
            return data

    class geom_abc(geom):
        DEFAULT_PARAMS = {"stat": stat_abc, "position": "identity"}
        REQUIRED_AES = {"x", "weight"}

        @staticmethod
        def draw(pinfo, panel_params, coord, ax, **kwargs):
            pass

    # weight is manually set, it should be a stat parameter and
    # not a geom manual setting
    g = geom_abc(weight=4)
    assert "weight" in g.aes_params
    assert "weight" in g._stat.params

    g = geom_abc(aes(weight="mpg"))
    assert "weight" in g.mapping
    assert "weight" in g._stat.params


def test_calculated_expressions():
    p = ggplot(mtcars, aes(x="factor(cyl)", y="..count..+1")) + geom_bar()
    # No exception
    p._build()


def test_removes_infinite_values():
    df = mtcars.copy()
    df.loc[[0, 5], "wt"] = [np.inf, -np.inf]
    p = ggplot(df, aes(x="wt")) + geom_bar()

    with pytest.warns(PlotnineWarning) as record:
        p._build()

    def removed_2_row_with_infinites(record):
        for item in record:
            msg = str(item.message).lower()
            if "2 rows" in msg and "non-finite" in msg:
                return True
        return False

    assert removed_2_row_with_infinites(record)



from __future__ import annotations

import typing

from .position import position

if typing.TYPE_CHECKING:
    from plotnine.typing import FloatArray, FloatArrayLike


class position_nudge(position):
    """
    Nudge points

    Useful to nudge labels away from the points
    being labels.

    Parameters
    ----------
    x : float
        Horizontal nudge
    y : float
        Vertical nudge
    """

    def __init__(self, x=0, y=0):
        self.params = {"x": x, "y": y}

    @classmethod
    def compute_layer(cls, data, params, layout):
        trans_x = None  # pyright: ignore
        trans_y = None  # pyright: ignore

        if params["x"]:

            def trans_x(x: FloatArrayLike) -> FloatArray:
                return x + params["x"]

        if params["y"]:

            def trans_y(y: FloatArrayLike) -> FloatArray:
                return y + params["y"]

        return cls.transform_position(data, trans_x, trans_y)



import pandas as pd

from plotnine import aes, geom_errorbar, geom_errorbarh, ggplot

n = 4
df = pd.DataFrame(
    {
        "x": [1] * n,
        "ymin": range(1, 2 * n + 1, 2),
        "ymax": range(2, 2 * n + 2, 2),
        "z": range(n),
    }
)


def test_errorbar_aesthetics():
    p = (
        ggplot(df, aes(ymin="ymin", ymax="ymax"))
        + geom_errorbar(aes("x"), size=2)
        + geom_errorbar(aes("x+1", alpha="z"), width=0.2, size=2)
        + geom_errorbar(aes("x+2", linetype="factor(z)"), size=2)
        + geom_errorbar(aes("x+3", color="z"), size=2)
        + geom_errorbar(aes("x+4", size="z"))
    )

    assert p == "errorbar_aesthetics"


def test_errorbarh_aesthetics():
    p = (
        ggplot(df, aes(xmin="ymin", xmax="ymax"))
        + geom_errorbarh(aes(y="x"), size=2)
        + geom_errorbarh(aes(y="x+1", alpha="z"), height=0.2, size=2)
        + geom_errorbarh(aes(y="x+2", linetype="factor(z)"), size=2)
        + geom_errorbarh(aes(y="x+3", color="factor(z)"), size=2)
        + geom_errorbarh(aes(y="x+4", size="z"))
    )

    assert p == "errorbarh_aesthetics"



from __future__ import annotations

import typing
from types import SimpleNamespace as NS
from typing import overload
from warnings import warn

from ..exceptions import PlotnineWarning
from ..iapi import panel_ranges, panel_view
from ..positions.position import transform_position
from .coord import coord, dist_euclidean

if typing.TYPE_CHECKING:
    from typing import Optional, Sequence

    import pandas as pd

    from plotnine.iapi import scale_view
    from plotnine.typing import (
        FloatArray,
        FloatSeries,
        Scale,
        Trans,
        TupleFloat2,
    )


class coord_trans(coord):
    """
    Transformed cartesian coordinate system

    Parameters
    ----------
    x : str | trans
        Name of transform or `trans` class to
        transform the x axis
    y : str | trans
        Name of transform or `trans` class to
        transform the y axis
    xlim : None | (float, float)
        Limits for x axis. If None, then they are
        automatically computed.
    ylim : None | (float, float)
        Limits for y axis. If None, then they are
        automatically computed.
    expand : bool
        If `True`, expand the coordinate axes by
        some factor. If `False`, use the limits
        from the data.
    """

    trans_x: Trans
    trans_y: Trans

    def __init__(
        self,
        x: str | Trans = "identity",
        y: str | Trans = "identity",
        xlim: Optional[TupleFloat2] = None,
        ylim: Optional[TupleFloat2] = None,
        expand: bool = True,
    ):
        from mizani.transforms import gettrans

        self.trans_x = gettrans(x)
        self.trans_y = gettrans(y)
        self.limits = NS(x=xlim, y=ylim)
        self.expand = expand

    def transform(
        self, data: pd.DataFrame, panel_params: panel_view, munch: bool = False
    ) -> pd.DataFrame:
        from mizani.bounds import squish_infinite

        if not self.is_linear and munch:
            data = self.munch(data, panel_params)

        def trans_x(col: FloatSeries) -> FloatSeries:
            result = transform_value(self.trans_x, col, panel_params.x.range)
            if any(result.isnull()):
                warn(
                    "Coordinate transform of x aesthetic "
                    "created one or more NaN values.",
                    PlotnineWarning,
                )
            return result

        def trans_y(col: FloatSeries) -> FloatSeries:
            result = transform_value(self.trans_y, col, panel_params.y.range)
            if any(result.isnull()):
                warn(
                    "Coordinate transform of y aesthetic "
                    "created one or more NaN values.",
                    PlotnineWarning,
                )
            return result

        data = transform_position(data, trans_x, trans_y)
        return transform_position(data, squish_infinite, squish_infinite)

    def backtransform_range(self, panel_params: panel_view) -> panel_ranges:
        return panel_ranges(
            x=self.trans_x.inverse(panel_params.x.range),
            y=self.trans_y.inverse(panel_params.y.range),
        )

    def setup_panel_params(self, scale_x: Scale, scale_y: Scale) -> panel_view:
        """
        Compute the range and break information for the panel
        """

        def get_scale_view(
            scale: Scale, coord_limits: TupleFloat2, trans: Trans
        ) -> scale_view:
            if coord_limits:
                coord_limits = trans.transform(coord_limits)

            expansion = scale.default_expansion(expand=self.expand)
            ranges = scale.expand_limits(
                scale.limits, expansion, coord_limits, trans
            )
            sv = scale.view(limits=coord_limits, range=ranges.range)
            sv.range = tuple(sorted(ranges.range_coord))
            sv.breaks = transform_value(
                trans,
                # TODO: fix typecheck
                sv.breaks,  # pyright: ignore
                sv.range,
            )
            sv.minor_breaks = transform_value(trans, sv.minor_breaks, sv.range)
            return sv

        out = panel_view(
            x=get_scale_view(scale_x, self.limits.x, self.trans_x),
            y=get_scale_view(scale_y, self.limits.y, self.trans_y),
        )
        return out

    def distance(
        self,
        x: pd.Series[float],
        y: pd.Series[float],
        panel_params: panel_view,
    ) -> FloatArray:
        max_dist = dist_euclidean(panel_params.x.range, panel_params.y.range)[
            0
        ]
        xt = self.trans_x.transform(x)
        yt = self.trans_y.transform(y)
        return dist_euclidean(xt, yt) / max_dist


@overload
def transform_value(
    trans: Trans, value: pd.Series[float], range: TupleFloat2
) -> pd.Series[float]:
    ...


@overload
def transform_value(
    trans: Trans, value: Sequence[float], range: TupleFloat2
) -> Sequence[float]:
    ...


def transform_value(
    trans: Trans, value: pd.Series[float] | Sequence[float], range: TupleFloat2
) -> pd.Series[float] | Sequence[float]:
    """
    Transform value
    """
    return trans.transform(value)
